/*
 * discrim.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <locale.h>
#include "robot-env.h"
#include "evonet.h"
#include "utilities.h"

int robottype = Khepera;			// the robot we are using
int  nfoods;						// number of food elements
double *robotsdist;                 // matrix containing robot distances;
int *robotsbydistance;              // matrix containing the id of the robots ordered by the inverse of the distance

/*
 * read parameters from the configuration file
 */
bool readEvoConfig(const char* filename)
{
    char *s;
    char buff[1024];
    char name[1024];
    char value[1024];
    char *ptr;
    int section;  // 0=before the section 1=in the section 2= after the section

    section = 0;

    FILE* fp = fopen(filename, "r");
    if (fp != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp) != NULL)
        {

            // check whether the relevant section start or end
            if (buff[0] == '[')
            {
              if (section == 1)
                {
                    section = 2;
                    continue;
                }
              if ((section == 0) && (strncmp(buff, "[EXP]",5)==0))
              {
                section = 1;
                continue;
              }
            }

            if (section == 1)
            {

            //Skip blank lines and comments
            if (buff[0] == '\n' || buff[0] == '#' || buff[0] == '/')
            continue;

            //Parse name/value pair from line
            s = strtok(buff, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, name);

            s = strtok(NULL, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, value);

            // Copy into correct entry in parameters struct
            if (strcmp(name, "nrobots")==0)
            nrobots = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "trials")==0)
            trials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "ttrials")==0)
            ttrials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steps")==0)
            steps = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steptimelength")==0)
            steptimelength = (int)strtol(value, &ptr, 10);
			else if (strcmp(name, "robottype")==0)
            robottype = (int)strtol(value, &ptr, 10);
            else printf("WARNING: Unknown parameter %s in section [EXP] of file %s \n", name, filename);
         }
        }
        fclose (fp);
        if (section == 0)
           printf("WARNING: Missing section [NET] in file %s \n", filename);

        printf("Robot %d nrobots %d Trials %d TTrials %d Steps %d\n", robottype, nrobots, trials, ttrials, steps);
        return(true);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
        return(false);
    }
}


/*
 * initialize the environment
 */
void
initEnvironment()

{

	int cobj=0;
	double wsize;
	double asize;
	int f;
	
    nfoods = 1;
	nenvobjs = 4 + nfoods;	// total number of objects
	switch (robottype)
	 {
	   case (Khepera):
	    wsize = 2000.0;
		asize = 120.0;
		break;
	   case (ePuck):
	    wsize = 2500.0;
		asize = 160.0;
		break;
	   case (MarXBot):
	    wsize = 5000.0;
		asize = 400.0;
		break;
	  }
	
	worldx = wsize;
	worldy = wsize;

	
	envobjs = (struct envobjects *) malloc(nenvobjs * sizeof(struct envobjects));
	
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = 0.0;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = 0.0;
	envobjs[cobj].y2 = worldy;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = worldx;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = worldy;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = worldy;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = worldy;
	cobj++;
	for(f=0; f < nfoods; f++)
	{
	  envobjs[cobj].type = STARGETAREA;
	  envobjs[cobj].x = 400.0;
	  envobjs[cobj].y = 400.0;
	  envobjs[cobj].r = asize;
	  envobjs[cobj].color[0] = 0.5;
	  cobj++;
	}
	
	if (cobj > nenvobjs)
		{
			printf("ERROR: you should allocate more space for environmental objects");
			fflush(stdout);
		}
	
}


/*
 * generate the initial states
 * if mode != 0 regenerate only the trial initial conditions
 */
void createInitialStates(int mode)

{
    
    int s;
    double *is;
	int numtrials;
	int attempts;
	double dx, dy;
	struct robot *ro1;
	struct robot *ro2;
	int r1, r2;
	double cdist, mindist;
	double fcx, fcy;
	double distfromborder;
	double distfromareacenter;
	
	if (mode == 0)
	  numtrials = trials + ttrials;
	 else
	  numtrials = trials;
	
	switch (robottype)
	 {
	   case (Khepera):
	    distfromborder = 300.0;
		distfromareacenter = 150.0;
		break;
	   case (ePuck):
	    distfromborder = 400.0;
		distfromareacenter = 200.0;
		break;
	   case (MarXBot):
	    distfromborder = 1000.0;
		distfromareacenter = 500.0;
		break;
	  }

    for (s=0, is = initialstates; s < numtrials; s++)
    {
      // home and foraging area
      fcx = *is = RNG->getDouble(distfromborder, worldx - distfromborder);
      is++;
      fcy = *is = RNG->getDouble(distfromborder, worldy - distfromborder);
      is++;
	  // initial positions and orientations of the robots
	  for (r1=0, ro1=rob; r1 < nrobots; r1++, ro1++)
	   {
        mindist = 0.0; attempts = 0;
		while(mindist < (ro1->radius*2+5) && attempts < 100)
		{
          ro1->dir = RNG->getDouble(0.0, PI2);
          ro1->x = fcx + RNG->getDouble(-distfromareacenter, distfromareacenter);
          ro1->y = fcy + RNG->getDouble(-distfromareacenter, distfromareacenter);
		  mindist = 99999;
	      for (r2=0, ro2=rob; r2 < r1; r2++, ro2++)
	       {
			  dx = (ro1->x - ro2->x);
			  dy = (ro1->y - ro2->y);
			  cdist = sqrt(dx*dx+dy*dy);
			  if (cdist < mindist)
				  mindist = cdist;
		   }
		 attempts++;
		}
	    *is = ro1->dir; is++;
		*is = ro1->x; is++;
		*is = ro2->y; is++;
	  }
    }
	

}

/*
 *  update the power-distances between robots and the matrix of nearest robots
 */
void updateRobotDistances()

{

    //double *robotsdist;     // matrix containing robot distances;
    //int *robotsbydistance;  // matrix containing the id of the robots ordered by the inverse of the distance

    struct robot *ro1;
    struct robot *ro2;
    int r1, r2, r3;
    double *cdist;
    double *ccdist;
    double smallerdist;
    int smallerid;
    int *rbd;
    double *rbddist;
    int remaining;


    // update the matrix of distances
    cdist = robotsdist;
    for (r1=0, ro1=rob; r1 < nrobots; r1++, ro1++)
    {
      for (r2=0, ro2=rob; r2 < r1; r2++, ro2++)
      {
        *cdist = ((ro1->x - ro2->x)*(ro1->x - ro2->x)) + ((ro1->y - ro2->y)*(ro1->y - ro2->y));
        *(robotsdist + (r1 + (r2 * nrobots))) = *cdist;
        cdist++;
      }
      *cdist = 9999999999.0;; // distance from itself is set to a large number to exclude itself from the nearest
      cdist = (cdist + (nrobots - r1));
    }

    // update the matrix of nearest robots
    rbd = robotsbydistance;
    for (r1=0, cdist = robotsdist; r1 < nrobots; r1++, cdist = (cdist + nrobots))
    {
      remaining = (nrobots - 1);
      for (r2=0; r2 < (nrobots - 1); r2++)
      {
        for (r3=0, ccdist = cdist, smallerid=0, smallerdist = *ccdist, rbddist = ccdist; r3 < nrobots; r3++, ccdist++)
        {
          if (*ccdist <= smallerdist)
          {
            smallerdist = *ccdist;
            smallerid = r3;
            rbddist = ccdist;
          }
        }
        if (smallerdist < 1000000)  // we ignore robots located at a power dstance greater that 750*750
         {
          *rbd = smallerid;
          rbd++;
          *rbddist = 9999999999.0;
          remaining--;
         }
         else
         {
          *rbd = -1;
          rbd = (rbd + remaining);
          break;
         }
      }
      *rbd = -1;  // we use -1 to indicate the end of the list
      rbd++;

    }

    /*
    cdist = robotsdist;
    for (r1=0; r1 < nrobots; r1++)
        {
          for (r2=0; r2 < nrobots; r2++)
             {
               printf("%10.1f ", *cdist);
               cdist++;
             }
             printf("\n");
         }
     printf("\n");
     */

    /*
    rbd = robotsbydistance;
    for (r1=0; r1 < nrobots; r1++)
        {
          for (r2=0; r2 < nrobots; r2++)
             {
               printf("%3d ", *rbd);
               rbd++;
             }
             printf("\n");
         }
    printf("\n");
    */

}



/*
 * initialize the environment, the robots, and the network
 * if mode != 0 regenerate the trial initial conditions
 * return the number of required parameters (-1 in case of failure)
 */
int initialize(const char* filename)

{
    
    struct robot *ro;
    struct evonet *ne;
	int r;


	// default values
    trials = 1;
    ttrials = 1;
    steps = 100;
	
	// set USA local conventions
	setlocale( LC_ALL, "en-US" );
	// read parameters from the .ini file
	if (readEvoConfig(filename) == false)
		return(-1);

    // create and configure the environment
    initEnvironment();

    // create the robots, the networks, and initialize the sensors
    rob = (struct robot *) malloc(nrobots * sizeof(struct robot));
	net = (struct evonet *) malloc(nrobots * sizeof(struct evonet));
    // initialize the robots and the networks
	for (r=0, ro=rob, ne=net; r < nrobots; r++, ro++, ne++)
	 {
	  // initialize the robots
      initRobot(ro, r, robottype);
	  // set maxspeed
	  switch (robottype)
	   {
	   case (Khepera):
	    ro->maxSpeed = 144.0;
		break;
	   case (ePuck):
        ro->maxSpeed = 200.0;
		break;
	   case (MarXBot):
        ro->maxSpeed = 500.0;
		break;
	   }
      ro->color = 4;  // frontal and rear side are red and blue, respectively
      // initialize sensors
      ne->ninputs = 0;
      ne->ninputs += initInfraredSensor(ro, 8, ne->ninputs);
      ne->ninputs += initCameraSensorRFB(ro, 2);
      ne->ninputs += initGroundSensor(ro);
      ne->ninputs += initEnergySensor(ro);
      ne->ninputs += initBiasSensor(ro);
      // initialize motors
      ro->motorwheels = 2;
      ro->motorwheelstype = 1; // encoding of speed and rotation
      ro->motorleds = 2;
      ne->noutputs = ro->motorwheels + ro->motorleds;
	  // allocate/initialize the input vector
	  initRobotSensors(ro, ne->ninputs);
      // initialize robot's network
      initNet(ne, filename, r);
      resetNet(ne);
      // set the id number of motors neurons
      ro->motorwheelsid = ne->ninputs + ne->nhiddens;
      ro->motorledsid = ne->ninputs + ne->nhiddens + ro->motorwheels;
	 }
	
	// generate initial states
    // (initial positions/orientations of robots and positions of foods area)
    initialstates = (double *) malloc(((trials+ttrials)*(3*nrobots+nfoods*2)) * sizeof(double));
    createInitialStates(0);

    // allocate distance matrices and initialize the dyagonal distances to 0
    robotsdist = (double *) malloc((nrobots*nrobots) * sizeof(double));
    robotsbydistance = (int *) malloc((nrobots*nrobots) * sizeof(int));

	nfreep = computeParameters(net);
	printf("Network parameters %d\n", nfreep);
	
	return(nfreep);
}

/*
 * evaluate a genotype for capability to perform the discrimination task
 * return the total fitness
 * when mode=1 evaluate a genotype on the test trials
 */
double evaluate(double* genotype, const int glen, int mode)

{
	
	int r,f,s;
    double *is;
    int trial;
	int numtrials;
    int step;
    double fitness;
	struct robot *ro;
	struct evonet *ne;
	bool stoptrial;
    double oldx, oldy; // updated every 2s
	double dx, dy;
    int foodarea[100];
	int collectedfood[100];
	double cdist;
    int *rbd;
    int x, y;
	
	
	cellsize = worldx / 20.0;
	
    // evaluation loop
	if (glen != nfreep)
	  {
	    printf("WARNING: The length of the parameter vector received by discriminate (%d) does not match the number of parameters required (%d)\n", glen, nfreep);
	  }
    fitness = 0.0;
	if (mode == 0)
	   {
	    is = initialstates;
		numtrials = trials;
		}
	  else
	   {
	    is = (initialstates + (trials * ((3 * nrobots) + (nfoods * 2))));
		numtrials = ttrials;
	   }
	for (trial=0; trial < numtrials; trial++)
    {
	    // the robots initially do not carry food
		for(r=0; r < nrobots; r++)
		 {
          foodarea[r] = 0;
		  collectedfood[r] = 0;
		 }
		stoptrial = false;

        for(x=0; x < 100; x++)
          for(y=0; y < 100; y++)
            cells[x][y] = 0;
		
	    for (r=0, ro=rob, ne=net; r < nrobots; r++, ro++, ne++)
		  {
           resetNet(ne);
			for(s=0,ro->csensors=ro->sensors; s < ne->ninputs; s++, ro->csensors++)
		      *ro->csensors = 0.0;
		  }
        // initialize the nest position
        for (f=0; f < nfoods; f++)
         {
          envobjs[f+4].x = *is; is++;
          envobjs[f+4].y = *is; is++;
         }
        // initialize robots position, orientation, and energy level
	    for (r=0, ro=rob; r < nrobots; r++, ro++)
	     {
           ro->alive = true;
		   ro->dir = *is; is++;
	       ro->x = *is; is++;
	       ro->y = *is; is++;
		   ro->energy = 1.0;
	     }
		//printf("trial %d robot1 %d %.2f %.2f %.2f robot2 %d %.2f %.2f %.2f \n", trial, rob->dir, rob->x, rob->y, b1, b2, b3);

        for (step=0; step < steps; step++)
        {
          for (r=0, ro=rob, ne=net, rbd = robotsbydistance; r < nrobots; r++, ro++, ne++, rbd = (rbd + nrobots))
		  {
            // update robots distances
            updateRobotDistances();
			// update sensors
            ro->csensors = ro->sensors;
            updateInfraredSensor(ro);
            updateCameraSensorRFB(ro, rbd);
			updateGroundSensor(ro);
            updateEnergySensor(ro);
			updateBiasSensor(ro);
			// update network
			updateNet(ne, ro->sensors, genotype);
			// update the robot
			oldx = ro->x;
			oldy = ro->y;
			updateRobot(ro, ne);
            // update robot energy
            dx = ro->x - envobjs[4].x;
            dy = ro->y - envobjs[4].y;
            cdist = sqrt((dx*dx)+(dy*dy));
			if (cdist < envobjs[4].r)
              {
               fitness += collectedfood[r];
			   collectedfood[r] = 0;
              }

            if (cdist < envobjs[4].r)
              {
               ro->energy = 1.0;
              }
            ro->energy -= 0.01;
            if (ro->energy < 0.0)
              {
                ro->alive = false;
              }
	
            // update fitness
            if (ro->energy > 0.0)
             {
             x = (int) ro->x / cellsize;
             y = (int) ro->y / cellsize;
             if (cells[x][y] == 0)
              {
                collectedfood[r] += 1;
                cells[x][y] = 1;
              }
			
			}
		  }
			


#ifdef __cplusplus
          if (renderrobotworld) update_world();		// update graphic rendering
#endif

			
	   }
		
	}
	
    return(fitness / numtrials);
    
}


