/*
 * co-evolutionary steady-state algorithm
 * require an evaluate function called evaluate2 which receive two genotypes for two corresponding individuals
 */



#include <stdlib.h>

#include <vector>
#include <algorithm>
#include <string.h>
#include <time.h>
#include <typeinfo>
#include <math.h>
#include "coevo.h"
#include "utilities.h"
#include <QVector>

#include "./Eigen/Core"
#include "./Eigen/Dense"
#include "./unsupported/Eigen/MatrixFunctions"

int trials=1;						// number of training trials
int ttrials=1;						// number of testing trials

int nsamples;						// number of sample vectors
double etaInd;						// Learning rate for the individual
double etaCovM;						// Learning rate for the covariance matrix
double *utility;					// array containing the utilities values
double stepRate = 1.0;				// the step rate

extern double complexlev[];         // this vector is defined and updated in predprey.cpp

double *bestggenotype;				// the best generalizing genotype

int popsize = 20;					// popsize (parents)
long int maxevaluations = 100;		// max number of evaluations
int seed = 1;						// random seed
int nreplications = 1;				// number of replications
double prange = 5.0;				// parameters range
double mutrate = 0.03;				// mutation rate
int steptimelength = 0;				// time length of a step during graphic visualization
int envchangeevery = 20;			// environment changed every n generations
int testgen=0;                      // generation to be tested
int testgeninterval=0;              // generation interval used during post-evaluation test
int evalonlyn=-1;                   // test only n individuals
int evalpred=-1;                    // test a specific predator
int evalprey=-1;                    // test a specific prey
int saveonlyeveryn=-1;              // save data only every n generations
bool savefitmatrix=false;           // save the fitness matrix
bool savecomplexitymatrix=false;    // save the complexity matrix
RandomGenerator* RNG;               // randomgenerator class


/*
 * read parameters from the configuration file
 */
bool readEvoParameters(const char* filename)
{
    char *s;
    char buff[1024];
    char name[1024];
    char value[1024];
    char *ptr;
    int section;  // 0=before the section 1=in the section 2= after the section

    section = 0;
    FILE* fp = fopen(filename, "r");
    if (fp != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp) != NULL)
        {
            // check whether the relevant section start or end
            if (buff[0] == '[')
            {
              if (section == 1)
                {
                    section = 2;
                    continue;
                }
              if ((section == 0) && (strncmp(buff, "[EVO]",5)==0))
              {
                section = 1;
                continue;
              }
            }

            if (section == 1)
            {

            //Skip blank lines and comments
            if (buff[0] == '\n' || buff[0] == '#' || buff[0] == '/')
            continue;

            //Parse name/value pair from line
            s = strtok(buff, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, name);

            s = strtok(NULL, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, value);

            // Copy into correct entry in parameters struct
            if (strcmp(name, "maxevaluations")==0)
            maxevaluations = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "nreplications")==0)
            nreplications = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "prange")==0)
            prange = strtod(value, &ptr);
            else if (strcmp(name, "mutrate")==0)
            mutrate = strtod(value, &ptr);
            else if (strcmp(name, "popsize")==0)
            popsize = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "seed")==0)
            seed = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steptimelength")==0)
            steptimelength = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "envchangeevery")==0)
            envchangeevery = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "testgen")==0)
            testgen = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "testgeninterval")==0)
            testgeninterval = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "evalonlyn")==0)
            evalonlyn = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "evalpred")==0)
            evalpred = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "evalprey")==0)
            evalprey = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "saveonlyeveryn")==0)
            saveonlyeveryn = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "savefitmatrix")==0)
            savefitmatrix = parsebool(value);
            else if (strcmp(name, "savecomplexitymatrix")==0)
            savecomplexitymatrix = parsebool(value);
            else printf("WARNING: Unknown parameter %s in section [EVO] of file %s \n", name, filename);
            }
         }
        fclose (fp);
        if (section == 0)
           printf("WARNING: Missing section [EVO] in file %s \n", filename);
        return(true);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
        return(false);
    }
}

/*
 * save a genotype in a file
 */
void saveGenotype(char* filename, double* genotype, int glen)

{

	FILE *fp;
	
	fp = fopen(filename,"wb!");
	if (!fp)
      printf("ERROR: unable to open file %s\n", filename);
     else
      {
      fwrite(&glen, sizeof(int), 1, fp);
	  fwrite(genotype, sizeof(double)*glen, 1, fp);
      }
	fclose(fp);

}

/*
 * load a genotype from a file
 */
int loadGenotype(char* filename, double* genotype, int glen)

{

	FILE *fp;
	size_t result;
    int np;
	
	fp = fopen(filename,"rb");
	if (!fp)
	 {
	  printf("ERROR: unable to open file %s\n", filename);
	  return(0);
	 }
     else
	 {
      fread(&np, sizeof(int), 1, fp);
      if (np != glen)
        {
         printf("ERROR: the number of parameters contained in the file %s differ from that expected\n", filename);
         //if (!feof(fp) || result != sizeof(double)*glen */)
         //  printf("ERROR: the number of parameters contained in the file %s differ from what is expected\n", filename);
        }
       else
        {
         result = fread(genotype, sizeof(double)*glen, 1, fp);
        }
      fclose(fp);
	 }
    return(1);
}

/*
 * save a population in a text file
 */
void savePopTxt(char* filename, double **pop, int glen, int popsize)

{

	FILE *fp;
	int p;
	int i;
	double *g;
	
	fp = fopen(filename,"w!");
	if (!fp)
      printf("ERROR: unable to open file %s\n", filename);
     else
      {
	  for(p=0; p < popsize; p++, pop++)
	   {
	    for(i=0, g = *pop; i < glen; i++, g++)
		  fprintf(fp, "%f ", *g);
	    fprintf(fp, "\n");
	   }
      }
	fclose(fp);

}


/*
 * save a population in a binary file
 */
void savePop(char* filename, double **pop, int glen, int popsize)

{

	FILE *fp;
	int p;
	
	fp = fopen(filename,"wb!");
	if (!fp)
      printf("ERROR: unable to open file %s\n", filename);
     else
      {
	  fwrite(&popsize, sizeof(int), 1, fp);
      fwrite(&glen, sizeof(int), 1, fp);
	  for(p=0; p < popsize; p++, pop++)
	    fwrite(*pop, sizeof(double)*glen, 1, fp);
      }
	fclose(fp);

}

/*
 * save two populations of popsize individuals ranked by fitness
 */
void saveRankedPops(char* filename, double **pop, int glen, int popsize, int *predrank, int *preyrank)

{

	FILE *fp;
	int p;
	int i;
	double *rgenome;
	int ntotal;
	
	ntotal = popsize*2;
	fp = fopen(filename,"wb!");
	if (!fp)
      printf("ERROR: unable to open file %s\n", filename);
     else
      {
	  fwrite(&ntotal, sizeof(int), 1, fp);
      fwrite(&glen, sizeof(int), 1, fp);
	  for(p=0; p < 2; p++)
       {
	     for(i=0; i < popsize; i++)
		 {
		   if (p == 0)
	         rgenome = *(pop + predrank[i]);
			else
			 rgenome = *(pop + popsize + preyrank[i]);
	      fwrite(rgenome, sizeof(double)*glen, 1, fp);
		 }
	   }
      }
	fclose(fp);

}

/*
 * load a genotype from a file
 */
void loadPop(char* filename, double **pop, int glen, int popsize)

{

	FILE *fp;
    int np;
	int ps;
	int p;
	
	fp = fopen(filename,"rb");
	if (!fp)
	 {
	  printf("ERROR: unable to open file %s\n", filename);
	 }
     else
	 {
	  fread(&ps, sizeof(int), 1, fp);
      fread(&np, sizeof(int), 1, fp);
      if (np != glen || ps != popsize)
        {
         printf("ERROR: the number of parameters contained in the file %s differ from that expected\n", filename);
        }
       else
        {
		 for(p=0; p < popsize; p++, pop++)
           fread(*pop, sizeof(double)*glen, 1, fp);
        }
      fclose(fp);
	 }
}


/*
 * copy and mutate
 */
void copyandmutate(double **genome, int ngenes, double mutr, double prange, int parent, int offspring)
{
    double **gpi;
    double **gpi2;
    double *gi;
    double *gi2;
    int g;
    
    gpi=(genome + parent);
    gpi2=(genome + offspring);
    
    for (g=0,gi = *gpi, gi2 = *gpi2; g < ngenes; g++, gi++, gi2++)
        {
            if (RNG->getDouble(0.0, 1.0) < mutr)
			 {
               if (RNG->getDouble(0.0, 1.0) < 0.5)
			     {
                  *gi2 = RNG->getDouble(-prange, prange);
			     }
			     else
			     {
                  *gi2 += *gi + RNG->getGaussian(1.0, 0.0);
				  if (*gi2 > prange)
				     *gi2 = prange;
				  if (*gi2 < -prange)
				     *gi2 = -prange;
				 }
			 }
             else
			 {
               *gi2 = *gi;
			 }
            
        }
}
/*
 * save the pop1*pop2 fitness matrix
 */
void savepopfit(int gen, double *fmatrix, double *p1fit, double *p2fit, int *p1rank, int *p2rank)
{

	char filen[256];
	FILE *fp;
	int i, c;
	double fitpop1, fitpop2;
    int *r1, *r2;
	double *f1, *f2;
	double *fm;
	double t1,t2;
	double predtot;
	double *preytot;
    double *ft;

    sprintf(filen, "G%dS%d.fit", gen, seed);
	fp = fopen(filen,"w!");
	
	preytot = (double *) malloc(popsize * sizeof(double));
	for(c=0, ft=preytot; c < popsize; c++, ft++)
	  *ft = 0.0;
	
    for(i=0, fm=fmatrix, fitpop1=0.0, fitpop2=0.0; i < popsize; i++)
      {
	    predtot = 0;
        for(c=0; c < popsize; c++, fm++)
         {
		   fprintf(fp, "%f ", *fm);
           fitpop1 += *fm;
           fitpop2 += 1.0 - *fm;
		   predtot += *fm;
		   *(preytot + c) += 1.0 - *fm;
         }
        fprintf(fp, "  %f\n", predtot / (double) popsize);
      };
	for(c=0, ft=preytot; c < popsize; c++, ft++)
	  fprintf(fp, "%f ", *ft / (double) popsize);
	fprintf(fp, "\n");
	
	fprintf(fp, "%dx%d average %f %f \n", popsize, popsize, fitpop1 / (double) (popsize*popsize), fitpop2 / (double) (popsize*popsize));

	fprintf(fp, "line P1Fit P1rank P1RankFit P2fit P2Rank P2RankFit\n");
	for(c=0, r1=p1rank, r2=p2rank, f1=p1fit, f2=p2fit, t1=0.0, t2=0.0; c < popsize; c++, r1++, r2++, f1++, f2++, t1 += *f1, t2 += *f2)
      fprintf(fp, "%d %f %d %f %f %d %f\n", c, *f1, *r1, *(p1fit + *r1), *f2, *r2, *(p2fit + *r2));
	fprintf(fp, "tot fit %f %f \n", t1, t2);
    fclose(fp);
}

/*
 * select the individual more similar to individual idc
 * the individuals not yet selected are listed inside the clusters list
 * the function return the cluster-id index which contain the id of the individuals
 */
int nearestCluster(int idc, double *fmatrix, QVector<QVector <int> > clusters, double *mdist)

{

   int c;
   double mindist = 999999.9;
   double dist;
   int nearest = 0;
   int i;
	
//printf("cluster idc %d size %d ", idc, clusters.size());
//for(int iii=0; iii < clusters.size(); iii++)
// printf("%d ", clusters[iii][0]);
//printf("\n");

   for(c=0; c < clusters.size(); c++)
   {
     //if (clusters[c][0] != idc)
     //{
      for (i=0, dist=0.0; i < popsize; i++)
       {
         dist += fabs(*(fmatrix + (idc * popsize + i)) - *(fmatrix + (clusters[c][0] * popsize + i)));
       }
      if (dist < mindist)
      {
       mindist = dist;
       nearest = c;
      }
     //}
   }

   *mdist = mindist;
   printf("(size %d) [%d %d] dist %.2f\n", clusters.size(), idc, clusters[nearest][0], mindist);
   for (i=0; i < 10; i++)
      printf("%.2f ", *(fmatrix + (idc * popsize + i)));
   printf("\n");
   for (i=0; i < 10; i++)
      printf("%.2f ", *(fmatrix + (clusters[nearest][0] * popsize + i)));
   printf("\n");
   //printf("return %d \n", nearest);
   return(nearest);

}

int nearestClusterBis(int idc, QVector<QVector <double> > fitmatrix, QVector<QVector <int> > clusters, double *mdist)

{

   int c;
   double mindist = 999999.9;
   double dist;
   int nearest = 0;
   int i;

   for(c=0; c < clusters.size(); c++)
   {
     if (clusters[c][0] != idc)
     {
      //printf("%d-%d ", c, clusters[c][0]);
      for (i=0, dist= 0.0; i < popsize; i++)
       {
         dist += fabs(fitmatrix[idc][i] - fitmatrix[clusters[c][0]][i]);
       }
      if (dist < mindist)
      {
       mindist = dist;
       nearest = c;
      }
     }
   }
   *mdist = mindist;
   //printf("\n");
   //printf("(size %d) [%d %d] dist %.2f\n", clusters.size(), idc, clusters[nearest][0], mindist);
   return(nearest);

}

/*
 * select the competitors by clustering them into 10 groups and by selecting the best of each group
 * select the evolving individuals by chosing the fittest against each selected competitor
 */
void select_individuals(int evopop, double *fmatrix, double *fitind0, double *fitind1, double **genome, double **selgenome, int gen, int ngenes, int *selectedl)

{

  int i,ii;
  int c;
  int ccc;
  int g;
  int ind;
  double fit;
  int cc;
  int randselected, selected;
  FILE *fp;
  char filen[256];
  double *ge;
  double *selge;
  int verbose = 1; // save cluster in Gn.fit file
  QVector<QVector <int> > clusters1;  // matrix which contains the id of the individual for each 1-element custer
  QVector<QVector <int> > clusters2;  // matrix that contains the id of the individuals for each 2-elements cluster
  QVector<QVector <int> > clusters3;  // matrix which contains the id of the individual for each 4-element custer
  QVector<QVector <int> > clusters4;  // matrix that contains the id of the individuals for each 8-elements cluster
  QVector<QVector <double> > fitmatrix2; // matrix containing the average fitness of 2 elements clusters
  QVector<QVector <double> > fitmatrix3; // matrix containing the average fitness of 4 elements clusters
  QVector<QVector <double> > fitmatrix4; // matrix containing the average fitness of 8 elements clusters
  QVector<int> clusterv; // vector of n elements to be added to a cluster matrix
  QVector<double> fitv; // vector of n elements to be added to a cluster fitness matrix
  QVector<int> competitors; // id of selected competitors
  double maxf;
  int besti;
  double max;
  int maxi;
  double dist;
  int c1, c2;
  double *mindist;
  double *dists;
  double *fmatrix2;
  double *fmatrix3;
  double *fmatrix4;
  double *f2, *f1i, *f2i;

   printf("Cluster competitors \n ");

   // distances among elements of each cluster
   dists = (double *)malloc(popsize * sizeof(double));
   mindist = (double *)malloc(1 * sizeof(double));

   // generate the initial clusters (containing groups of one individuals)
   clusters1.resize(popsize);
   for(c=0; c < popsize; c++)
    {
     clusters1[c].resize(1);
     clusters1[c][0] = c;
    }

   // generate the first set of clusters (containing groups of two elements)
   // the first element of each group is selected randomly, the second is the nearest among the remaining
   clusters2.resize(0);
   clusterv.resize(3);

   fitmatrix2.resize(0);
   fitv.resize(popsize);
   cc = 0;
   *mindist = 0.0;
   while(clusters1.size() > 1)
    {
     clusterv[0] = cc;
     randselected = RNG->getInt(0, clusters1.size()-1);
     clusterv[1] = clusters1[randselected][0];
     f1i = (fmatrix + (clusterv[1] * popsize));
     for(i=0; i < popsize; i++, f1i++) fitv[i] = *f1i / 2.0;
     clusters1.remove(randselected);
     selected = nearestCluster(clusterv[1], fmatrix, clusters1, mindist);
     clusterv[2] = clusters1[selected][0];
     *(dists + cc) = *mindist;
     f1i = (fmatrix + (clusterv[2] * popsize));
     for(i=0; i < popsize; i++, f1i++) fitv[i] += *f1i / 2.0;
     clusters2.append(clusterv);
     clusters1.remove(selected);
     fitmatrix2.append(fitv);
     cc++;
	}

   if (verbose > 0)
   {
    sprintf(filen, "G%dS%d.fit", gen, seed);
    fp = fopen(filen,"a");
    for(i=0, mindist = dists; i < 40; i++, mindist++)
       fprintf(fp, "cluster %d [%d %d] %.2f \n", clusters2[i][0], clusters2[i][1], clusters2[i][2], *mindist);
    for(i=0; i < fitmatrix2.size(); i++)
     {
      for(ii=0; ii < popsize; ii++, f1i++)
        fprintf(fp, "%.2f ", fitmatrix2[i][ii]);
      fprintf(fp, "\n");
     }
   }

   // generate the second set of clusters (containing groups of four elements)
   // the first element of each group is selected randomly, the second is the nearest among the remaining
   clusters3.resize(0);
   clusterv.resize(5);
   fitmatrix3.resize(0);
   fitv.resize(popsize);
   cc = 0;
   while(clusters2.size() > 1)
    {
     clusterv[0] = cc;
     for(i=0; i < popsize; i++) fitv[i] = 0.0;
     randselected = RNG->getInt(0, clusters2.size()-1);
     for(ccc=0; ccc < 2; ccc++)
        {
         clusterv[ccc+1] = clusters2[randselected][ccc+1];
         for(i=0, f1i = (fmatrix + (clusterv[ccc+1] * popsize)); i < popsize; i++, f1i++) fitv[i] += *f1i / 4.0;
        }
     clusters2.remove(randselected);
     selected = nearestClusterBis(randselected, fitmatrix2, clusters2, mindist);
     *(dists + cc) = *mindist;
     for(ccc=0; ccc < 2; ccc++)
        {
         clusterv[ccc+3] = clusters2[selected][ccc+1];
         for(i=0, f1i = (fmatrix + (clusterv[ccc+3] * popsize)); i < popsize; i++, f1i++) fitv[i] += *f1i / 4.0;
        }
     clusters3.append(clusterv);
     clusters2.remove(selected);
     fitmatrix3.append(fitv);
     cc++;
    }

   if (verbose > 0)
   {
    for(i=0, mindist = dists; i < 20; i++, mindist++)
       fprintf(fp, "cluster %d [%d %d %d %d] %.2f\n", clusters3[i][0], clusters3[i][1], clusters3[i][2], clusters3[i][3], clusters3[i][4], *mindist);
    for(i=0; i < 20; i++)
     {
      for(ii=0; ii < popsize; ii++)
        fprintf(fp, "%.2f ", fitmatrix3[i][ii]);
      fprintf(fp, "\n");
     }
   }

   // generate the fourth set of clusters (containing groups of 8 elements)
   // the first element of each group is selected randomly, the second is the nearest among the remaining
   clusters4.resize(0);
   clusterv.resize(9);
   fitmatrix4.resize(0);
   fitv.resize(popsize);
   cc = 0;
   while(clusters3.size() > 1)
    {
     clusterv[0] = cc;
     for(i=0; i < popsize; i++) fitv[i] = 0.0;
     randselected = RNG->getInt(0, clusters3.size()-1);
     for(ccc=0; ccc < 4; ccc++)
        {
         clusterv[ccc+1] = clusters3[randselected][ccc+1];
         for(i=0, f1i = (fmatrix + (clusterv[ccc+1] * popsize)); i < popsize; i++, f1i++) fitv[i] += *f1i / 8.0;
        }
     clusters3.remove(randselected);
     selected = nearestClusterBis(randselected, fitmatrix3, clusters3, mindist);
     *(dists + cc) = *mindist;
     for(ccc=0; ccc < 4; ccc++)
        {
         clusterv[ccc+5] = clusters3[selected][ccc+1];
         for(i=0, f1i = (fmatrix + (clusterv[ccc+5] * popsize)); i < popsize; i++, f1i++) fitv[i] += *f1i / 8.0;
        }
     clusters4.append(clusterv);
     clusters3.remove(selected);
     fitmatrix4.append(fitv);
     cc++;
    }

   if (verbose > 0)
   {
    for(i=0, mindist = dists; i < 10; i++, mindist++)
    {
       fprintf(fp, "cluster %d [", clusters4[i][0]);
       for(ii=0; ii < 8; ii++)
         fprintf(fp, "%d ", clusters4[i][1+ii]);
       fprintf(fp, "] %.2f\n", *mindist);
    }
   }




   // calculate similarity between and within clusters
   if (verbose > 0)
   {
   long int ncomp;
   int iii;
   fprintf(fp, "clusters within distances ");
   for(i=0; i < clusters4.size(); i++)
    {
     for(c1=1, dist = 0.0, ncomp = 0; c1 < clusters4[i].size(); c1++)
     {
	   for(c2=c1+1; c2 < clusters4[i].size(); c2++)
        {
		  f1i = (fmatrix + (clusters4[i][c1] * popsize));
		  f2i = (fmatrix + (clusters4[i][c2] * popsize));
		  for(ii=0; ii < popsize; ii++, ncomp++, f1i++, f2i++)
		    dist += fabs(*f1i - *f2i);
		  //printf("cluster %d AxB %d %d (%d %d) dist %.2f \n", i, c1, c2, clusters4[i][c1], clusters4[i][c2], dist);
		}
	 }
	 fprintf(fp, "%.3f ", dist / (double) ncomp);
	}
	fprintf(fp, "\n");
	
    fprintf(fp, "clusters between distances\n");
	// 10 clusters x 10 clusters x 8 individuals x 8 individuals x 80 values
    for(c1=0; c1 < clusters4.size(); c1++)
     {
	  for(c2=0; c2 < clusters4.size(); c2++)
	   {
		   dist = 0.0, ncomp = 0;
		   for(i=1; i < clusters4[i].size(); i++)
            {
	          for(ii=1; ii < clusters4[i].size(); ii++)
			  {
			     f1i = (fmatrix + (clusters4[c1][i] * popsize));
			     f2i = (fmatrix + (clusters4[c2][ii] * popsize));
				 for(iii=0; iii < popsize; iii++, ncomp++, f1i++, f2i++)
		           dist += fabs(*f1i - *f2i);
				 //printf("cluster %d %d elem AxB %d %d (%d %d) dist %.2f \n", c1, c2, i, ii, clusters4[c1][i], clusters4[c2][ii], dist);
			  }
			}
		   fprintf(fp, "%.3f ", dist / (double) ncomp);
		}
		fprintf(fp, "\n");
	  }
   }
	
	
   // for each cluster we indentify the fittest individual and ww copy its genotype in the selected population
   printf("Gen %d Evopop %d Selected competitors ", gen, evopop);
   competitors.resize(clusters4.size());
   for(i=0; i < clusters4.size(); i++)
   {
     for(ii=1, maxf=0.0, besti=0; ii < clusters4[i].size(); ii++)
     {
       ind = clusters4[i][ii];
       if (evopop == 0)
         fit = *(fitind1 + clusters4[i][ii]);
        else
         fit = *(fitind0 + clusters4[i][ii]);
       if (fit > maxf)
        {
          besti = ind;
          maxf = fit;
        }
     }
     competitors[i] = besti;
     printf("%d ", besti);
     if (evopop == 0)
      {
       ge = *(genome + besti + popsize);
       selge = *(selgenome + i + 10*2);
	   *(selectedl + i + 10) = besti;
      }
      else
      {
       ge = *(genome + besti);
       selge = *(selgenome + i);
	   *(selectedl + i) = besti;
      }
     for(g=0; g < ngenes; g++, ge++, selge++)
      {
       *selge = *ge;
      }
   }
   printf("\n");
	
   // select the evolving individuals randomly
   QVector <int> indexes;
   indexes.resize(popsize);
   int s;
   for(i=0; i < popsize; i++)
     indexes[i] = i;
   printf("Gen %d Evopop %d Selected evolving ind ", gen, evopop);
   for(i=0; i < 10; i++)
	{
       s = RNG->getInt(0, indexes.size()-1);
	   ind = indexes[s];
	   indexes.remove(s);
	   printf("%d ", ind);
       if (evopop == 0)
        {
          ge = *(genome + ind);
          selge = *(selgenome + i);
		  *(selectedl + i) = ind;
        }
       else
        {
         ge = *(genome + ind + popsize);
         selge = *(selgenome + i + 10*2);
		 *(selectedl + i + 10) = ind;
        }
	  for(g=0; g < ngenes; g++, ge++, selge++)
        {
         *selge = *ge;
		}
   }
   printf("\n");
	
   /* we know select the evolving individuals with the highest fitness against the selected competitors
   printf("Gen %d Evopop %d Selected evolving ind ", gen, evopop);
   for(c=0; c < competitors.size(); c++)
   {
       for(i=0, max=0.0, maxi=0; i < popsize; i++)
       {
	      f1i = (fmatrix + (i * popsize) + c);
          if (*f1i > max)
          {
            max = *f1i;
            maxi = i;
          }
       }
       printf("%d ", maxi);
       if (evopop == 0)
        {
          ge = *(genome + maxi);
          selge = *(selgenome + c);
		  *(selectedl + i) = maxi;
        }
       else
        {
         ge = *(genome + maxi + popsize);
         selge = *(selgenome + c + 10*2);
		 *(selectedl + i + 10) = maxi;
        }
	  for(g=0; g < ngenes; g++, ge++, selge++)
        {
         *selge = *ge;
		}
   }
   printf("\n"); */


   fclose(fp);
}


/*
 * select the competitors by choosing well differentiated individual
 * select the evolving individuals by chosing them randomly
 */
void select_individuals_differentiated(int evopop, double **genome, double **selgenome, int gen, int ngenes, int *selected, double *fmatrix)

{

  double *dists;
  int re;
  double dist;
  int c;
  double *m1;
  double *m2;
  int sel;
  int ss;
  int *sc;
  int g;
  double maxdist;
  double *di;
  int s;
  int i;
  int *selcompetitors;
  double *ge;
  double *selge;
  int ind;

   dists = (double *)malloc(popsize * sizeof(double));
   selcompetitors = (int *)malloc(10 * sizeof(int));

   // select the competitors individuals
   printf("Gen %d Evopop %d Selected differentiated competitors ", gen, evopop);
   QVector <int> indexes;
   indexes.resize(popsize);
   for(i=0; i < popsize; i++)
     indexes[i] = i;
   // the first individual is selected randomly
   sel = RNG->getInt(0, indexes.size()-1);
   for(s=0, sc=selcompetitors; s < 10; s++, sc++)
    {
      *sc = indexes[sel];
      indexes.remove(sel);
      printf("%d ", *sc);
      if (evopop == 0)
       {
        ge = *(genome + *sc + popsize);
        selge = *(selgenome + s + 10*2);
        *(selected + s + 10) = *sc;
       }
       else
       {
        ge = *(genome + *sc);
        selge = *(selgenome + s);
        *(selected + s) = *sc;
       }
      for(g=0; g < ngenes; g++, ge++, selge++)
       {
        *selge = *ge;
       }
      // we now compute the distance of each remaining element
	  if (s < (10 - 1))
      for(re=0; re < indexes.size(); re++)
        {
//printf("selected ");
//for(int sz=0; sz <= s; sz++)
// printf("%d ", *(selcompetitors + sz));
//printf("remaining %d \n", indexes[re]);
          for(ss=0, dist=0; ss <= s; ss++)
           {
              m1 = (fmatrix + (indexes[re] * popsize));              // the current remaining individual
              m2 = (fmatrix + (*(selcompetitors + ss) * popsize));   // the current selected individual
              {
               for(c=0; c < popsize; c++, m1++, m2++)
			     {
                  dist += fabs(*m1 - *m2);
				  //if (evopop == 1) printf("%.2f ", *m1);
				 }
			   //if (evopop == 1) printf("\n");
              }
           }
//printf("dist %.2f \n", dist);
        *(dists + indexes[re]) = dist;
        //printf("%d dist %.2f\n", re, dist);
        }
      // we select the more distant remaining individual
      for(re=0, maxdist=0; re < indexes.size(); re++)
        {
          di = (dists + indexes[re]);
          if (*di > maxdist)
          {
            maxdist = *di;
            sel = re;
          }
        }
       //printf("the more distant is %d dist %.2f \n", sel, maxdist);
       }
    printf("\n");

   //for(re=0; re < indexes.size(); re++)
      //printf("%d dist %.2f\n", indexes[re], *(dists + indexes[re]));

   // select the evolving individuals randomly
   indexes.resize(popsize);
   for(i=0; i < popsize; i++)
     indexes[i] = i;
   printf("Gen %d Evopop %d Selected evolving ind ", gen, evopop);
   for(i=0; i < 10; i++)
    {
       s = RNG->getInt(0, indexes.size()-1);
       ind = indexes[s];
       indexes.remove(s);
       printf("%d ", ind);
       if (evopop == 0)
        {
          ge = *(genome + ind);
          selge = *(selgenome + i);
          *(selected + i) = ind;
        }
       else
        {
         ge = *(genome + ind + popsize);
         selge = *(selgenome + i + 10*2);
         *(selected + i + 10) = ind;
        }
      for(g=0; g < ngenes; g++, ge++, selge++)
        {
         *selge = *ge;
        }
   }
   printf("\n");

}

/*
 * select the competitors and the evolving individuals randomly
 */
void select_individuals_random(int evopop, double **genome, double **selgenome, int gen, int ngenes, int *selected)

{
	
  int ind;
  int s;
  int i;
  int g;
  double *ge;
  double *selge;
	


  // select the predators
  printf("Gen %d Evopop %d Selected predators ", gen, evopop);
  QVector <int> indexes;
  indexes.resize(popsize);
  for(i=0; i < popsize; i++)
    indexes[i] = i;
  for(i=0; i < 10; i++)
   {
     s = RNG->getInt(0, indexes.size()-1);
	 ind = indexes[s];
	 indexes.remove(s);
	 printf("%d ", ind);
	 *(selected + i) = ind;
	 ge = *(genome + ind);
	 selge = *(selgenome + i);
     for(g=0; g < ngenes; g++, ge++, selge++)
       *selge = *ge;
   }
   printf("\n");
	
   // select the prey
   indexes.resize(popsize);
   for(i=0; i < popsize; i++)
     indexes[i] = i;
   printf("Gen %d Evopop %d Selected prey ind ", gen, evopop);
   for(i=0; i < 10; i++)
	{
       s = RNG->getInt(0, indexes.size()-1);
	   ind = indexes[s];
	   indexes.remove(s);
	   printf("%d ", ind);
	   *(selected + i + 10) = ind;
	   ge = *(genome + ind + popsize);
	   selge = *(selgenome + i + 10*2);
	   for(g=0; g < ngenes; g++, ge++, selge++)
         *selge = *ge;
   }
   printf("\n");
   
   
	
	
  /* select the competitors individuals
  printf("Gen %d Evopop %d Selected competitors ", gen, evopop);
  QVector <int> indexes;
  indexes.resize(popsize);
  for(i=0; i < popsize; i++)
    indexes[i] = i;
  for(i=0; i < 10; i++)
   {
     s = RNG->getInt(0, indexes.size()-1);
	 ind = indexes[s];
	 indexes.remove(s);
	 printf("%d ", ind);
     if (evopop == 0)
      {
       ge = *(genome + ind + popsize);
       selge = *(selgenome + i + 10*2);
	   *(selected + i + 10) = ind;
      }
      else
      {
       ge = *(genome + ind);
       selge = *(selgenome + i);
	   *(selected + i) = ind;
      }
     for(g=0; g < ngenes; g++, ge++, selge++)
      {
       *selge = *ge;
      }
   }
   printf("\n");
	
   // select the evolving individuals
   indexes.resize(popsize);
   for(i=0; i < popsize; i++)
     indexes[i] = i;
   printf("Gen %d Evopop %d Selected evolving ind ", gen, evopop);
   for(i=0; i < 10; i++)
	{
       s = RNG->getInt(0, indexes.size()-1);
	   ind = indexes[s];
	   indexes.remove(s);
	   printf("%d ", ind);
       if (evopop == 0)
        {
          ge = *(genome + ind);
          selge = *(selgenome + i);
		  *(selected + i) = ind;
        }
       else
        {
         ge = *(genome + ind + popsize);
         selge = *(selgenome + i + 10*2);
		 *(selected + i + 10) = ind;
        }
	  for(g=0; g < ngenes; g++, ge++, selge++)
        {
         *selge = *ge;
		}
   }
   printf("\n");
   */
   
}


static double xNesStep(double *geno, std::vector<double*> competitors, Eigen::MatrixXd& matrix, int ngenes, int nsteps, int nsamples, double * utility)
{

    double fit;
    double bestfit;
    int s;
    int i;
    int j;
    double *g;
    Eigen::VectorXd ind(ngenes); // Copy of the individual
    Eigen::VectorXd dInd(ngenes); // The update of the individual
    Eigen::MatrixXd dCovM(ngenes,ngenes); // The update of the covariance matrix
    Eigen::MatrixXd expCovM(ngenes,ngenes); // Exponential matrix of CovM
    Eigen::MatrixXd samples(ngenes,nsamples); // Samples
    Eigen::MatrixXd trSamples(ngenes,nsamples); // Transposed samples
    Eigen::MatrixXd offspring(ngenes,nsamples); // Offspring
    Eigen::MatrixXd eye(ngenes,ngenes); // Identity matrix
    double *genotype = (double*)malloc(sizeof(double) * ngenes);
    double *samfit = (double*)malloc(sizeof(double) * nsamples);
    double *ordf = (double*)malloc(sizeof(double) * nsamples);
    int *ordoff = (int*)malloc(sizeof(int) * nsamples);
    double ordUtilSum;
    double* orderedUtility = (double*)malloc(sizeof(double) * nsamples);
	
    // Copy the individual (TO BE FIXED)
    for (i = 0, g = geno; i < ngenes; i++, g++)
        ind[i] = *g;
	
    // Initialize the identity matrix
    for (i = 0; i < ngenes; i++)
    {
        eye(i,i) = 1.0;
        for (j = 0; j < ngenes; j++)
        {
            if (i != j)
                eye(i,j) = 0.0;
        }        
    }
    bestfit = -9999.0;
	
    // Perform <nsteps> steps of the xNES algorithm
    for (s = 0; s < nsteps; s++)
    {
        // Calculate the matrix exponential
        Eigen::MatrixExponential<Eigen::MatrixXd>(matrix).compute(expCovM);
		
        // Draw out samples from a Gaussian distribution 
        for (i = 0; i < ngenes; i++)
        {
            for (j = 0; j < nsamples; j++)
                samples(i,j) = RNG->getGaussian(1.0, 0.0);
        }
        // Create the offspring
        trSamples = expCovM * samples;
        for (j = 0; j < nsamples; j++)
        {
            for (i = 0; i < ngenes; i++)
                offspring(i,j) = ind[i] + trSamples(i,j);
        }
        // Evaluate offspring
        for (j = 0; j < nsamples; j++)
        {
            for (i = 0; i < ngenes; i++)
                genotype[i] = offspring(i,j);
            samfit[j] = evaluate(genotype, ngenes, 0);
	    }
        // Sorting samples based on their fitness
        sortSamples(samfit, nsamples, ordf, ordoff, true);
		
        // Now fill the array with the utility values
        for (j = 0, ordUtilSum = 0.0; j < nsamples; j++)
        {
            orderedUtility[ordoff[j]] = utility[j];
            ordUtilSum += orderedUtility[ordoff[j]];
        }
        // Compute the natural gradient for both the individual and the covariance matrix
        // Create a matrix whose rows contain weights * samples
        Eigen::MatrixXd ordUtil(ngenes,nsamples);
        Eigen::MatrixXd ordUtilSamples(ngenes,nsamples);
        Eigen::MatrixXd Z(nsamples,ngenes);
        for (i = 0; i < ngenes; i++)
        {
            for (j = 0; j < nsamples; j++)
                ordUtil(i,j) = orderedUtility[j];
        }
        for (i = 0; i < ngenes; i++)
        {
            for (j = 0; j < nsamples; j++)
            {
                ordUtilSamples(i,j) = ordUtil(i,j) * samples(i,j);
                Z(j,i) = samples(i,j);
            }
        }
        Eigen::MatrixXd G = ordUtilSamples * Z - ordUtilSum * eye;
        // Define transpose matrices for both samples and ordered utilities
        // (they are necessary to perform the calculation of gradients)
        Eigen::MatrixXd ordUtilTr(nsamples,1);
        Eigen::MatrixXd v(ngenes,1);
        for (j = 0; j < nsamples; j++)
            ordUtilTr(j,0) = orderedUtility[j];
        v = etaInd * expCovM * samples * ordUtilTr;
        for (i = 0; i < nsam; i++)
            dInd[i] = v(i,0);
        dCovM = etaCovM * G;
        // Update the individual and the covariance matrix
        for (i = 0; i < ngenes; i++)
        {
            ind[i] += (stepRate * dInd[i]);
        }
        matrix += (stepRate * dCovM);
        // Evaluate the individual
        for (i = 0; i < ngenes; i++)
            genotype[i] = ind[i];
        fit = evaluate(genotype, ngenes, 0);
        if (fit > ordf[0])
        {
            if (fit > bestfit)
                bestfit = fit;
        }
        else
        {
            if (ordf[0] > bestfit)
                bestfit = ordf[0];
        }
        /*printf("Step %d - bestfit %lf\n", s, bestfit);
        fflush(stdout);*/
    }
    // Update the individual
    for (i = 0, g = geno; i < ngenes; i++, g++)
        *g = ind[i];
    return bestfit;
}

/*
 * Competitive co-evolutionary algorithm
 */
void evolve(int ngenes)
{

	int changevery = 100;				// change evolving pop every n generations
    double **genome;
    double **selgenome;
    double *fit;
    double f;
    double *fi;
	double *fip;
	double *fio;
    double **gpi;
	double **gpic;
    double *gi;
    double tx;
    int ceval;
    int cgen;
	double maxgf;
    int i;
    int j;
	int c;
	int e;
	int g;
	int p;
	int re, rp;
    char filen[256];
    double *statistics;                     // performance across generations
    double *stat;                           // performance across generations, local pointer
    long int statn;                         // number of stat data
	//double mutr;							// mutation rate (reduced to 0.01 during last phase)
    int evopop;                             // the evolving population (0 or 1)
	int startpred, startprey;
    double fitpop1, fitpop2;                // average fitness of population 1 and 2
	double fitsel1, fitsel2;				// average fitness of selected individuals of pop1 and pop2
	double fitsel;
	double fitselgain, fitpopgain;			// the fitness gain of selected individuals and of the overall population
	double fitselinitial;                   // the initial fitness of selected individuals against selected competitors
	double *fmatrix;						// fitness matrix
	double *fmatrix2;						// fitness matrix trasposed
	double *fmatrixevo;						// fitness matrix of evolving individuals tested agaist all competitors
	double *predfit;
    double *preyfit;
    double *evofit;
	double *pfit;
	int *selected;
	int *prank;
    int *predrank;
    int *preyrank;
    int *evorank;
    double *p1, *p2;
	double *ge;
	double *selge;
	double minf;
	double *fm;
    double *fme;
    double *fmp;
	double *cfit;
	FILE *fp;
	
	double u;
	double utilSum;
	
    int row;
    int col;
    std::vector<double*> competitors(popsize);
    std::vector<Eigen::MatrixXd> covMatrices(popsize);
    int startTime;
    int endTime;
    int elapsedTime;
    int sTime;
    int eTime; 
 
	printf("Co-Evolve-XNES: Seed: %d MaxEval %ld popsize %d prange %.1f\n", seed, maxevaluations, popsize, prange);
	
	// set the seed and generate the initial states
    RNG->setSeed(seed);
    createInitialStates(0);
	
	// Initialize parameters for the xNES
    nsamples = (int) (4.0 + floor(3.0 * log((double) ngenes)));
    etaInd = 1.0;
    etaCovM = (9.0 + 3.0 * log((double) ngenes)) / (5.0 * (double) ngenes * sqrt((double) ngenes));
    utility = (double*) malloc(sizeof(double) * nsamples);
    utilSum = 0.0;
    for (j = 0; j < nsamples; j++)
    {
        u = log(nsamples / 2.0 + 1.0) - log((double)j + 1.0);
        if (u < 0.0) u = 0.0;
        utility[j] = u;
        utilSum += u;
    }
    if (utilSum != 0.0)
	  for (j = 0; j < nsamples; j++) utility[j] /= utilSum;
	
	// allocate and initialize genome for two populations
    // a single matrix in which the second population is stored after the first population
    genome = (double **) malloc(popsize*2 * sizeof(double *));
    for (i=0, gpi=genome; i < popsize*2; i++,gpi++)
    {
        *gpi = (double *)malloc(ngenes * sizeof(double));
        for (j=0,gi = *gpi; j < ngenes; j++,gi++)
        {
            *gi = RNG->getDouble(-prange, prange);
        }
    }
	
    // Create and initialize covariance matrix for each individual in the population
    for (i = 0; i < popsize*2; i++)
    {
     covMatrices[i].resize(ngenes,ngenes);
     for (row = 0; row < ngenes; row++)
      {
        for (col = 0; col < ngenes; col++)
            covMatrices[i](row,col) = 0.0;
      }
    }
	
	// allocate and initialize selected genome (parent and offspring)
    // genome are stored in this order, parent-pop1, offspring-pop1, parent-pop2, offspring-pop2
	// only the offspring of the current evolving population are generated and eventually used to replace their parents
    selgenome = (double **) malloc(10*4 * sizeof(double *));
    for (i=0, gpi=selgenome; i < 10*4; i++,gpi++)
    {
        *gpi = (double *)malloc(ngenes * sizeof(double));
        for (j=0,gi = *gpi; j < ngenes; j++,gi++)
        {
            *gi = RNG->getDouble(-prange, prange);
        }
    }

    // allocate the indfit vector used to store the fitness of the evolving individuals (parents and offspring)
    fit = (double *) malloc(10*2 * sizeof(double));


    for (i = 0, gpi=genome, fi=fit; i < popsize; i++, gpi++, fi++)
    {
    *fi = xNesStep(*gpi, competitors, covMatrices[i], ngenes, 1000);
    printf("Individual %d - fitness %lf - elapsedTime %d\n", i, *fi, (eTime - sTime));
    fflush(stdout);
    }

    endTime = time(NULL);
    elapsedTime = endTime - startTime;
    printf("evolve() method took %d seconds\n", elapsedTime);
    fflush(stdout);
	

    // allocate memory for statistics
    statn = (maxevaluations / (trials * popsize) * 7) + 100;
    statistics = (double *) malloc(statn * sizeof(double));
    stat = statistics;
    statn = 0;
	// initialize fitness statistics
	maxgf = 0.0;
	

    // allocate the vectors to store the fitness and the ranking of pop1 and pop2
	// the fitness of the prey is inverted (high values are good for both predators and prey)
	fmatrix = (double *) malloc(popsize*popsize * sizeof(double));
	fmatrix2 = (double *) malloc(popsize*popsize * sizeof(double));
	fmatrixevo = (double *) malloc(10*popsize * sizeof(double));
    preyfit = (double *) malloc(popsize * sizeof(double));
    predfit = (double *) malloc(popsize * sizeof(double));
    predrank = (int *) malloc(popsize * sizeof(int));
    preyrank = (int *) malloc(popsize * sizeof(int));
	selected = (int *) malloc(10*2 * sizeof(int));  // selected predators and prey
    // allocate the vector to store the fitness of the evolving individuals against all competitore
    evofit = (double *) malloc(10 * sizeof(double));
    evorank = (int *) malloc(popsize * sizeof(int));
	
	// allocate memory for statistics
    statn = (maxevaluations / (10 * 10) * 4) + 100;  // number of generations
    statistics = (double *) malloc(statn * sizeof(double));
    stat = statistics;
    statn = 0;


	ceval = 0;
	cgen = 0;
	// we start by evolving predators against prey competitors
	// the evopop variable is initially set to 1 but is switched to 0 before the initial generation
	evopop = 1;
	
	// evaluate populations
	printf("Gen %d Evaluate %d against %d individuals\n", cgen, popsize, popsize);
	for(c=0,p1=predfit,p2=preyfit; c < popsize; c++, p1++, p2++)
		*p1 = *p2 = 0.0;
	for(i=0, gpi=genome, fm=fmatrix, fitpop1=0.0, fitpop2=0.0, p1=predfit; i < popsize; i++, gpi++, p1++)
               {
                 for(c=0, gpic=(genome + popsize), tx=0.0; c < popsize; c++, gpic++, fm++)
                  {
					*fm = f = evaluate2(*gpi, *gpic, ngenes, 0);
                    ceval++;
                    fitpop1 += f;
                    fitpop2 += 1.0 - f;
                    tx += f;
                    *p1 += f;
                    *(preyfit + c) += (1.0 - f); // prey fitness is the inverse of predator fitness
                  }
               }
	for(c=0,p1=predfit,p2=preyfit; c < popsize; c++, p1++, p2++)
		{
		   *p1 = *p1 / (double) popsize;
		   *p2 = *p2 / (double) popsize;
		}
	sortDoubles(popsize, predfit, predrank);
	sortDoubles(popsize, preyfit, preyrank);
	printf("Average PopFit: %.2f %.2f\n", fitpop1 / (double) (popsize*popsize), fitpop2 / (double) (popsize*popsize));
	savepopfit(cgen, fmatrix, predfit, preyfit, predrank, preyrank);
    sprintf(filen, "G0S%d.bgen", seed);
    savePop(filen, genome, ngenes, popsize*2);
    sprintf(filen, "G0S%dranked.bgen", seed);
    saveRankedPops(filen, genome, ngenes, popsize, predrank, preyrank);
	
    while (ceval <= maxevaluations)
      {

		// the current evolving phase ends and the new phase begin
        if ((cgen % changevery) == 0)
		   {

             if (cgen > 0)
             {
			  // evaluate the evolved individuals against all competitors
              printf("Gen %d Evopop %d Gen ", cgen, evopop);
				 
              if (evopop == 0) // evolved predators against all prey
			    {
                 for(i=0, gpi=selgenome, fi=evofit, fm = fmatrixevo; i < 10; i++, gpi++, fi++)
                  {
                    for(c=0, gpic=(genome + popsize), *fi = 0.0; c < popsize; c++, gpic++, fm++)
                    {
					  *fm = evaluate2(*gpi, *gpic, ngenes, 0);
					  ceval++;
					  *fi += *fm;
                    }
                   *fi = *fi / (double) popsize;
                   printf("%.2f ", *fi);
				  }
                 printf("\n");
               }
               else
               { // evolved prey against all predators
                 for(i=0, gpic=(selgenome + 10*2), fi=evofit, fm = fmatrixevo; i < 10; i++, gpic++, fi++)
                 {
                    for(c=0, gpi=genome, *fi = 0.0; c < popsize; c++, gpi++, fm++)
                    {
					   *fm = evaluate2(*gpi, *gpic, ngenes, 0);
					   ceval++;
                       *fi += 1.0 - *fm;
                    }
                   *fi = *fi / (double) popsize;
                   printf("%.2f ", *fi);
                 }
               }

               // replace the worst pop individuals with the evolved individuals displaying better performance against all competitors
			   // and update fitmatrix
			   printf("Gen %d Evopop %d Replaced \n", cgen, evopop);
			   sortDoubles(10, evofit, evorank);
			   fitpopgain = 0.0;
			   e = 10 - 1;  // we start by trying to replace the worst population individual with the worst evolved individual
			   if (evopop == 0)
			      {
			       pfit = predfit;
				   prank = predrank;
				  }
			     else
				  {
			      pfit = preyfit;
				  prank = preyrank;
				  }
			   while (e >= 0)
		        {
				  re = *(evorank + e);
				  for(p = (popsize - 1), rp = popsize - 1, minf = *(pfit + (popsize - 1)); p >= 0; p--) // find the current worst
				    {
				     if (*(pfit + p) <= minf)
					   {
					     rp = p;
						 minf = *(pfit + p);
					   }
					 }
			      if (*(evofit + re) >= *(pfit + rp))
				   {
				    printf("%d->%d(%.2f %.2f) ", rp, re,  *(pfit + rp), *(evofit + re));
					fitpopgain = *(evofit + re) - *(pfit + rp);
					if (evopop == 0)
					 {
					  ge = *(genome + rp);
					  selge = *(selgenome + re);
					 }
					 else
					 {
					  ge = *(genome + popsize + rp);
                      selge = *(selgenome + 10*2 + re);
					 }
					// replace genes
					for(g=0; g < ngenes; g++, ge++, selge++)
                       *ge = *selge;
					// update the fitness matrix and the fitness of the evolving population,
					if (evopop == 0)
					 {
					  *(predfit + rp) = *(evofit + re);
					  fme = (fmatrixevo + re*popsize);
					  fmp = (fmatrix + rp*popsize);
					  for(i=0; i < popsize; i++, fmp++, fme++)
					    *fmp = *fme;
					 }
					 else
					  {
					   *(preyfit + rp) = *(evofit + re);
					   fme = (fmatrixevo + re*popsize);
					   for(i=0; i < popsize; i++, fme++)
					    {
						  *(fmatrix + rp + (i * popsize)) = *fme;
						}
					    *fmp = *fme;
					  }
				   }
                  e--; // we now process the next better evolving individual
			    }
				printf("\n");
				// and we recalculate the fitness of the competitors on the basis of the fitness matrix
				// the arrival of new individuals in one population in fact affect also the fitness of the other population
				if (evopop == 0)
				  {
					for(i=0, cfit = preyfit; i < popsize; i++, cfit++)
					   *cfit = 0.0;
				    for(i=0, fm=fmatrix; i < popsize; i++)
				      for(c=0; c < popsize; c++, fm++)
					     {
						  *(preyfit + c) += 1.0 - *fm;
						 }
					for(i=0, cfit = preyfit; i < popsize; i++, cfit++)
					   *cfit = *cfit / (double) popsize;
				  }
				  else
				  {
					for(i=0, fm=fmatrix, pfit=predfit; i < popsize; i++, pfit++)
				      for(c=0, *pfit = 0.0; c < popsize; c++, fm++)
						  *pfit += *fm;
					for(i=0, pfit = predfit; i < popsize; i++, pfit++)
					  *pfit = *pfit / (double) popsize;
				  }
/*cd ..
double *f1;
double *f2;
printf("replaced evopop %d\n", evopop);
for(c=0, f1=predfit, f2=preyfit; c < popsize; c++, f1++, f2++)
  printf("%f %f\n", *f1, *f2);
*/
				// update the ranking of the evolving population
				if (evopop == 0)
				  sortDoubles(popsize, predfit, predrank);
			     else
	              sortDoubles(popsize, preyfit, preyrank);
                if (saveonlyeveryn <= 0 || (cgen % saveonlyeveryn) == 0)
                {
                  savepopfit(cgen, fmatrix, predfit, preyfit, predrank, preyrank);
                  sprintf(filen, "G%dS%d.bgen", cgen, seed);
                  savePop(filen, genome, ngenes, popsize*2);
                  sprintf(filen, "G%dS%dranked.bgen", cgen, seed);
                  saveRankedPops(filen, genome, ngenes, popsize, predrank, preyrank);
                }
			  } // end if cgen > 0
			  
			 // switch the evolving and competing populations
			 if (evopop == 0)
                evopop = 1;
			  else
                evopop = 0;
			
             // select competitors and evolving individuals for the new phase
			 //select_individuals_random(evopop, genome, selgenome, cgen, ngenes, selected);
			 
			
			if (evopop == 0)
             {
				 
               for(i=0, fi=fmatrix; i < popsize; i++)
               {
                 for(c=0; c < popsize; c++, fi++)
                   *(fmatrix2 + (i * popsize) + c) = *fi;
               }
			   select_individuals_differentiated(evopop, genome, selgenome, cgen, ngenes, selected, fmatrix2);
				 
             }
             else
             {
			  select_individuals_differentiated(evopop, genome, selgenome, cgen, ngenes, selected, fmatrix);
             }
			 

             /*
             if (evopop == 0)
             {
				 
               for(i=0, fi=fmatrix; i < popsize; i++)
               {
                 for(c=0; c < popsize; c++, fi++)
                   *(fmatrix2 + (i * popsize) + c) = *fi;
               }
               select_individuals(evopop, fmatrix2, predfit, preyfit, genome, selgenome, cgen, ngenes, selected);
				 
             }
             else
             {
              select_individuals(evopop, fmatrix, predfit, preyfit, genome, selgenome, cgen, ngenes, selected);
             }
			 */
			 
			 
			   
			 
             //evaluate the fitness of selected evolving individuals against selected competitors
//sprintf(filen, "SelG%dS%d.gen", cgen, seed);							// DEBUG
//savePopTxt(filen, selgenome, ngenes, 10*4);
//sprintf(filen, "G%dS%d.gen", cgen, seed);							    // DEBUG
//savePopTxt(filen, genome, ngenes, popsize*2);
//for(i=0; i < 20; i++)
//  printf("%d ", *(selected + i));
//printf("\n");
/*
for(i=0; i < 10; i++)
	for(c=0; c < 10; c++)
	   {
	   gpi = (genome + *(selected + i));
	   gpic = (genome + *(selected + 10 + c));
	   f = evaluate2(*gpi, *gpic, ngenes, evopop);
	   }
*/
             printf("Gen %d EvoPop %d Fit ", cgen, evopop);
			 for(i=0, fi = fit, fitsel = 0.0; i < 10; i++, fi++)
			   *fi = 0.0;
             for(i=0, gpi=selgenome; i < 10; i++, gpi++)
               {
                 for(c=0, gpic=(selgenome + 10*2); c < 10; c++, gpic++)
		          {
					f = evaluate2(*gpi, *gpic, ngenes, evopop);
					//printf("%f ", f);
					if (evopop == 0)
					  {
					   if (f != *(fmatrix + (*(selected + i) * popsize) + *(selected + 10 + c)))
                         printf("ERROR: Sanity check failed, retest of individuals %d %d produced a different fitness %f %f\n",
                                i,c, f,  *(fmatrix + (*(selected + i) * popsize) + *(selected + 10 + c)));
					  }
					  else
					  {
					   if ((1.0 - f) != *(fmatrix + (*(selected + i) * popsize) + *(selected + 10 + c)))
                         printf("ERROR: Sanity check failed, retest of individuals %d %d produced a different fitness %.2f %.2f\n",
                               i, c, (1.0 - f), *(fmatrix + (*(selected + i) * popsize) + *(selected + 10 + c)));
					  }
					fitsel += f;
					if (evopop == 0)
					  *(fit + i) += f;
				     else
					  *(fit + c) += f;
		            ceval++;
		          }
				  //printf("\n");
               }
			 for(i=0, fi = fit; i < 10; i++, fi++)
			   {
			    *fi = *fi / 10.0;
			    printf("%.2f ", *fi);
			   }
			 fitsel = fitsel / (10*10);
			 printf("(%.2f)\n", fitsel);
			 fitselinitial = fitsel;
		   }
		   else
           {
            // generate offspring
            for (i=0; i < 10; i++)
			  if (evopop == 0)
                copyandmutate(selgenome, ngenes, mutr, prange, i, i + 10);
			   else
			    copyandmutate(selgenome, ngenes, mutr, prange, 10*2 + i, 10*2 + 10 + i);
			
            // evaluate offspring
			if (evopop == 0)
			 {
		      startpred = 10;
			  startprey = 0;
			 }
			 else
			 {
		      startpred = 0;
			  startprey = 10;
			 }
			for(i=0, fi = (fit + 10); i < 10; i++, fi++)
			   *fi = 0.0;
            for(i=0, gpi=(selgenome + startpred), fitsel1=0.0; i < 10; i++, gpi++)
             {
              for(c=0, gpic=(selgenome + 10*2 + startprey); c < 10; c++, gpic++)
               {
			     f = evaluate2(*gpi, *gpic, ngenes, evopop);
				 if (evopop == 0)
				    *(fit + 10 + i) += f;
				  else
				    *(fit + 10 + c) += f;
                 ceval++;
               }
			 }
			 for(i=0, fi = (fit + 10); i < 10; i++, fi++)
			    *fi = *fi / 10.0;
			 // replace parents with their offspring when the former is outperformed by the latter
			for(i=0, fip = fit, fio = (fit + 10); i < 10; i++, fip++, fio++)
			  {
			    if (*fio >= *fip)
				 {
				   *fip = *fio;
			       if (evopop == 0)
				       copyandmutate(selgenome, ngenes, 0.0, prange, i + 10, i);
					else
				      copyandmutate(selgenome, ngenes, 0.0, prange, 10*2 + 10 + i, 10*2 + i);
				 }
			   }
			 // display the current fitness of evolving individuals
			 fitselgain = 0.0;
             printf("Gen %d EvoPop %d Fit ", cgen, evopop);
             for(i=0, fi=fit; i < 10; i++, fi++)
			   {
                printf("%.2f ", *fi);
				fitselgain += *fi;
			   }
			 fitselgain = (fitselgain / 10.0) - fitselinitial;
			 printf(" (%.2f %.2f) \n", fitselinitial, fitselgain);
			}

          // store stat data
          if ((cgen % changevery) == 0)
		   {
		    for(i=0, fi = predfit, fio = preyfit, fitpop1=0.0, fitpop2=0.0; i < popsize; i++, fi++, fio++)
			  {
			    fitpop1 += *fi;
				fitpop2 += *fio;
			  }
            *stat = fitpop1; stat++;  // pop1 average fitness
            *stat = fitpop2; stat++;  // pop2 average fitness
			*stat = fitselgain; stat++;  // fitness gain of selected individuals against selected competitors
            statn++;
		   }
		  
		  cgen += 1;
      }

     sprintf(filen, "statS%d.fit", seed);
	 int s,ss;
     if  ((fp = fopen(filen,"w!")) != NULL)
         {
            fprintf(fp, "gen fitpo1 fitpop2 fitselgain\n");
            for(s=0, stat = statistics; s < statn; s++)
              {
               for(ss=0; ss < 3; ss++, stat++)
                 fprintf(fp, "%lf ", *stat);
               fprintf(fp, "\n");
               }
            fclose(fp);
         }

	 // save the population
	//sprintf(filen,"pop.bgen");
    //savePop(filen, genome, ngenes, popsize*2);

 }


/*
 * meta evaluate function which call evaluate2 multiple times with different genotypes
 */
double evaluate(double *bestgenotype, int ngenes, int mode)
{

    double **genome;
	double **gpi;
	double **gpic;
	char filen[256];
	int i,ii;
	double f;
	
    // allocate and initialize genome (parent and offspring) for two populations
	// genome are stored in this order, parent-pop1, offspring-pop1, parent-pop2, offspring-pop2
    genome = (double **) malloc(popsize*2 * sizeof(double *));
    for (i=0, gpi=genome; i < popsize*2; i++,gpi++)
        *gpi = (double *)malloc(ngenes * sizeof(double));
	
    if (evalpred >= 0 && evalpred < popsize && evalprey >= 0 && evalprey < popsize)
    {
      printf("Evaluate: Seed %d Gen %d Individuals %d %d\n",seed, testgen, evalpred, evalprey);
      sprintf(filen,"G%dS%d.bgen", testgen, seed);
      loadPop(filen, genome, ngenes, popsize*2);
      gpi=(genome + evalpred);
      gpic=(genome + popsize + evalprey);
      f = evaluate2(*gpi, *gpic, ngenes, 0);
      printf("%.2f\n", f);
    }
    else
    {
     if (evalonlyn > 0)
      printf("Evaluate: Seed %d Gen %d Best %dx%d individuals\n",seed, testgen, evalonlyn, evalonlyn);
     else
      printf("Evaluate: Seed %d Gen %d (ranked) \n",seed, testgen);
        
     sprintf(filen,"G%dS%dranked.bgen", testgen, seed);
     loadPop(filen, genome, ngenes, popsize*2);
    
	 for(i=0, gpi=genome; i < popsize; i++, gpi++)
	  {
	   for(ii=0, gpic=(genome + popsize); ii < popsize; ii++, gpic++)
		  {
			 f = evaluate2(*gpi, *gpic, ngenes, 0);
			 printf("%.2f ", f);
             if (evalonlyn > 0 && ii == (evalonlyn - 1)) break;
		  }
        if (evalonlyn > 0 && i == (evalonlyn - 1)) break;
		printf("\n");
	  }
    }
  return(0.0);

 }

/*
 * save the fit matrix and average individual performance
 * mode = 0 (fitness) mode = 1 (complexity)
 */
void save_fitnessmatrix(double *fitm, int gen1, int gen2)
{
    
    double *f;
    int i, ii;
    FILE *fp;
    char fname[256];
    double af[2][popsize];
    double tav;
    
    if (gen1 == gen2)
        sprintf(fname, "S%dG%d.fit", seed, gen1);
    else
        sprintf(fname, "S%dG%dG%d.fit", seed, gen1, gen2);

    fp = fopen(fname, "w!");
    
    for (i=0, tav=0.0; i < popsize; i++)
        af[0][i] = af[1][i] = 0.0;
    
    for (f=fitm, i=0; i < popsize; i++)
    {
      for (ii=0; ii < popsize; ii++, f++)
      {
         fprintf(fp, "%f ", *f);
         af[0][i]  += *f;
         af[1][ii] += *f;
         tav += *f;
      }
      fprintf(fp, "\n");
    }
    
    fprintf(fp, "predators average ");
    for (i=0; i < popsize; i++)
        fprintf(fp, "%.2f ", af[0][i] / (double) popsize);
    fprintf(fp, "\n");
    fprintf(fp, "prey average ");
    for (i=0; i < popsize; i++)
        fprintf(fp, "%.2f ", af[1][i] / (double) popsize);
    fprintf(fp, "\n");
    fprintf(fp, "total average: %.3f\n", tav / (double) (popsize*popsize));
    
    fclose(fp);
    
}

/*
 * save the complexity matrix
 */
void save_complexitymatrix(double *matrix, int gen1, int gen2)
{
    
    double *f;
    int i, ii;
    FILE *fp;
    char fname[256];
    double af[2][popsize];
    int naf[2][popsize];
    int ntav;
    double tav;
    int p;
    

    if (gen1 == gen2)
         sprintf(fname, "S%dG%d.cpx", seed, gen1);
        else
         sprintf(fname, "S%dG%dG%d.cpx", seed, gen1, gen2);

    fp = fopen(fname, "w!");
    
    f = matrix;
    
    for(p=0; p < 2; p++)
    {
     for (i=0, tav=0.0, ntav=0; i < popsize; i++)
     {
        af[0][i] = af[1][i] = 0.0;
        naf[0][i] = naf[1][i] = 0;
     }
     for (i=0; i < popsize; i++)
     {
        for (ii=0; ii < popsize; ii++, f++)
        {
            fprintf(fp, "%f ", *f);
            if (*f > 0)
            {
             af[0][i]  += *f;
             af[1][ii] += *f;
             tav += *f;
             naf[0][i]  += 1;
             naf[1][ii] += 1;
             ntav++;
            }
        }
        fprintf(fp, "\n");
     }
     fprintf(fp, "lines average ");
     for (i=0; i < popsize; i++)
        fprintf(fp, "%.2f ", af[0][i] / (double) naf[0][i]);
     fprintf(fp, "\n");
    
     fprintf(fp, "columns average ");
     for (i=0; i < popsize; i++)
        fprintf(fp, "%.2f ", af[1][i] / (double) naf[1][i]);
     fprintf(fp, "\n");
     fprintf(fp, "total average: %.3f\n", tav / (double) ntav);
    }
    
    fclose(fp);
    
}
 
/*
 * postevaluate populations
 */
void test(int ngenes)
{

    double **genome1;
    double **genome2;
    double **gpi;
    double **gpic;
    double *fitmatrix;
    double *complexitymatrix;
    double *fm;
    double *cm;
    double *cm2;
    double f;
    char filen[256];
    int i,ii;
    int c;
    double totf;
    FILE *fp;
    int mastermode=0; // 0=only a generation, 1=against current and previous generation, 2= against previous and future generation

    printf("Test: Seed %d Gen %d Interval %d SaveFitmatrix %d SaveComplexitymatrix %d\n",seed, testgen, testgeninterval,savefitmatrix, savecomplexitymatrix);
           
    // allocate the genome for two populations
    // genome are stored in this order, parent-pop1, offspring-pop1, parent-pop2, offspring-pop2
    genome1 = (double **) malloc(popsize*2 * sizeof(double *));
    for (i=0, gpi=genome1; i < popsize*2; i++,gpi++)
        *gpi = (double *)malloc(ngenes * sizeof(double));
    genome2 = (double **) malloc(popsize*2 * sizeof(double *));
    for (i=0, gpi=genome2; i < popsize*2; i++,gpi++)
        *gpi = (double *)malloc(ngenes * sizeof(double));
    if (savefitmatrix)
      fitmatrix = (double *) malloc(popsize*popsize * sizeof(double));
    if (savecomplexitymatrix)
        complexitymatrix = (double *) malloc(popsize*popsize*2 * sizeof(double));

    // open result file
    sprintf(filen,"masterS%d.fit", seed);
    fp = fopen(filen, "a");

    // test predator of testgen against predator of the same and (eventually) previous generations
    if (testgeninterval == 0)
       {
        testgeninterval = testgen * 2 + 1; // if the interval parameter is 0 we only test against competitors of the same generation
        fprintf(fp, "Performace of Gen %d \n", testgen);
       }
       else
       {
         mastermode = 1;
         fprintf(fp, "Performace Predators of Gen %d against progressively ancient competitors (every %d)\n", testgen,testgeninterval);
       }
    sprintf(filen,"G%dS%d.bgen", testgen, seed);
    loadPop(filen, genome1, ngenes, popsize*2);
    for(c=testgen; c >= 0; c -= testgeninterval)
      {
        sprintf(filen,"G%dS%d.bgen", c, seed);
        loadPop(filen, genome2, ngenes, popsize*2);
          if (savefitmatrix) fm = fitmatrix;
          if (savecomplexitymatrix) {cm = complexitymatrix; cm2 = (complexitymatrix + (popsize*popsize));}
        for(i=0, gpi=genome1, totf=0.0; i < popsize; i++, gpi++)
         {
           for(ii=0, gpic=(genome2 + popsize); ii < popsize; ii++, gpic++)
              {
                 totf += f = evaluate2(*gpi, *gpic, ngenes, 0);
                 if (savefitmatrix) {*fm = f; fm++; }
                 if (savecomplexitymatrix) {*cm = complexlev[0]; *cm2 = complexlev[1]; cm++, cm2++;}
              }
         }
       fprintf(fp, "%.2f ", totf / (double) (popsize*popsize));
       printf("%.2f ", totf / (double) (popsize*popsize));
       if (!savefitmatrix) fprintf(fp, "savefitmatrix false \n");
       if (savefitmatrix) save_fitnessmatrix(fitmatrix, testgen, c);
       if (savefitmatrix) save_complexitymatrix(complexitymatrix, testgen, c);
      }
    fprintf(fp, "\n");
    printf("\n");
    // test prey of testgen against predator of the same and previous generations
    if (mastermode > 0)
    {
     fprintf(fp, "Performace Prey of Gen %d against progressively ancient competitors (every %d)\n", testgen, testgeninterval);
     sprintf(filen,"G%dS%d.bgen", testgen, seed);
     loadPop(filen, genome1, ngenes, popsize*2);
     for(c=testgen; c >= 0; c -= testgeninterval)
      {
        sprintf(filen,"G%dS%d.bgen", c, seed);
        loadPop(filen, genome2, ngenes, popsize*2);
        if (savefitmatrix) fm = fitmatrix;
        if (savecomplexitymatrix) {cm = complexitymatrix; cm2 = (complexitymatrix + (popsize*popsize));}
        for(i=0, gpi=genome2, totf=0.0; i < popsize; i++, gpi++)
         {
           for(ii=0, gpic=(genome1 + popsize); ii < popsize; ii++, gpic++)
              {
                 totf += f = evaluate2(*gpi, *gpic, ngenes, 1);
                 if (savefitmatrix) {*fm = f; fm++; }
                 if (savecomplexitymatrix) {*cm = complexlev[0]; *cm2 = complexlev[1]; cm++, cm2++;}
              }
         }
       fprintf(fp, "%.2f ", totf / (double) (popsize*popsize));
       printf("%.2f ", totf / (double) (popsize*popsize));
       if (savefitmatrix) save_fitnessmatrix(fitmatrix, c, testgen);
       if (savefitmatrix) save_complexitymatrix(complexitymatrix, c, testgen);
       }
     fprintf(fp, "\n");
     printf("\n");
     }

    fclose(fp);
 }






