/*
 * createsample.cpp
 * This utility experiment can be used to generate sample files
 * It operates by making the robot turns 360 degrees at different distances from
 *  an object and by storing the state of the simulated sensors in .sample files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "robot-env.h"
#include "evonet.h"
#include "utilities.h"

int ninputs;             // number of imput units

/*
 * read parameters from the configuration file
 */
bool readEvoConfig(const char* filename)
{
    char *s;
    char buff[1024];
    char name[1024];
    char value[1024];
    char *ptr;
    int section;  // 0=before the section 1=in the section 2= after the section

    section = 0;

    FILE* fp = fopen(filename, "r");
    if (fp != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp) != NULL)
        {

            // check whether the relevant section start or end
            if (buff[0] == '[')
            {
              if (section == 1)
                {
                    section = 2;
                    continue;
                }
              if ((section == 0) && (strncmp(buff, "[EXP]",5)==0))
              {
                section = 1;
                continue;
              }
            }

            if (section == 1)
            {

            //Skip blank lines and comments
            if (buff[0] == '\n' || buff[0] == '#' || buff[0] == '/')
            continue;

            //Parse name/value pair from line
            s = strtok(buff, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, name);

            s = strtok(NULL, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, value);

            // Copy into correct entry in parameters struct
            if (strcmp(name, "trials")==0)
            trials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "ttrials")==0)
            ttrials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steps")==0)
            steps = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steptimelength")==0)
            steptimelength = (int)strtol(value, &ptr, 10);
            else printf("WARNING: Unknown parameter %s in section [EXP] of file %s \n", name, filename);
         }
        }
        fclose (fp);
        if (section == 0)
           printf("WARNING: Missing section [EXP] in file %s \n", filename);

        printf("Trials %d TTrials %d Steps %d\n", trials, ttrials, steps);
        return(true);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
        return(false);
    }
}


/*
 * save an obstacle in a .sam file
 */
void
save_obstacle(char *filename, int **object, int  *objectcf)

{

   int  **ob;
   int  *o;
   FILE *fp;
   int  i,v,s;

   // open the file
   if ((fp = fopen(filename,"w")) == NULL)
	  {
	    printf("I cannot open file %s", filename);
	    fflush(stdout);
	  }

   // write the object configuration data:
   //      0-nsensors, 1-nangles, 2-ndistances,3-initdist,4-distinterval
   for (i=0, o=objectcf; i < 5; i++, o++)
     fprintf(fp,"%d ",*o);
   fprintf(fp,"\n");
	
   // write sample data
   for (v=0, ob=object; v < objectcf[2]; v++, ob++)
    {
     fprintf(fp,"TURN %d\n",v);
     for (i = 0, o = *ob; i < objectcf[1]; i++)
      {
       for (s=0; s < objectcf[0]; s++,o++)
        {
	      fprintf(fp,"%f ", *o / (float) 1023.0);
         }
       fprintf(fp,"\n");
      }
    }
   fprintf(fp, "END\n");
   fclose(fp);
}

/*
 * load an obstacle from a .sam file (oldstyle with integers)
 */
int **
load_obstacle_old(char *filename, int  *objectcf)

{

   int **object;
   int  **ob;
   int  *o;
   FILE *fp;
   int  i,ii,t,v,s;
   char sbuffer[128];

   if ((fp = fopen(filename,"r")) == NULL)
	 {
	    printf("ERROR: I cannot open file %s", filename);
		fflush(stdout);
	  }
   // read configuration data: 0-nsensors, 1-nangles, 2-ndistances,
   //                          3-initdist,4-distinterval
   for (i=0, o=objectcf; i < 5; i++, o++)
     fscanf(fp,"%d ",o);
   fscanf(fp,"\n");
	
   // allocate space and initialize
   object = (int **) malloc(objectcf[2] * sizeof(int *));
   ob = object;
   for (i=0; i<objectcf[2]; i++,ob++)
	  {
	   *ob = (int *)malloc((objectcf[0] * objectcf[1]) * sizeof(int));
	   for (ii=0,o=*ob; ii < (objectcf[0] * objectcf[1]); ii++)
		*o = 0;
	  }
   // load data
   for (t=0, ob=object; t < objectcf[2]; t++, ob++)
    {
     fscanf(fp,"TURN %d\n",&v);
     for (i = 0, o = *ob; i < objectcf[1]; i++)
      {
       for (s=0; s < objectcf[0]; s++,o++)
        {
	   fscanf(fp,"%d ",o);
        }
       fscanf(fp,"\n");
      }
    }
   // sanity check
   fscanf(fp, "%s\n",sbuffer);
   if (strcmp(sbuffer,"END") != 0)
     {
      printf("loading file %s: the number of sample is inconsistent", filename);
	  fflush(stdout);
	 }
	
   fclose(fp);
   return object;
}

/*
 *	update the laser sensor
 *  measure the closest distance detected in each perceptual sector
 */
int initSampleLaserDistanceSensor(struct robot *cro)
{
   if (cro->nlasersensors <= 0 || cro->nlasersensors > 8)
     cro->nlasersensors = 8;
   printf("Sensor[%d]: laser sensors\n", cro->nlasersensors);
   return(cro->nlasersensors);
}

/*
 * verify whether the angle a is in the range [r1,r2]
 * assume that a,r1 and r2 are in the range [0, PI2]
 */
bool anginside2(double a, double r1, double r2)

{

    if ((r2 - r1) > 0)
    {
      if ((r2 - r1) < M_PI)
        {
         if (a > r1 && a < r2) // clockwise from r1 to r2
           return(true);
          else
           return(false);
        }
         else
        {
         if (a > r2 || a < r1) // clockwise from r2 to PI2 and from 0 to r1
           return(true);
          else
           return(false);
         }
    }
    else
    {
        if ((r1 - r2) < M_PI)
          {
           if (a > r2 && a < r1) // counter-clockwise from r2 to r1
             return(true);
            else
             return(false);
          }
           else
          {
           if (a > r1 || a < r2) // couner-clockwise from r1 to PI2 and from 0 to r2
             return(true);
            else
             return(false);
           }
    }


}

void updateSampleLaserDistanceSensor(struct robot *cro)

{

    double maxdist = 105;          // maximum dist at which  objects are detected
    double starta = -M_PI;          // initial angle of the first sectonr
    double inta = M_PI / (double) (cro->nlasersensors / 2);
    int ob;
    double x, y, dx, dy, dist;
    double a1, a2;
    double angle;                   // angle of cylindrical object or of the nearest point of a wall
    double angle1, angle2;          // angle of the initial and final point of a wall
    int s,r;
    struct robot *ro;
    double mdist[8];                // minimum detected distance
    double cdist;
    double ldist;                   // lateral distance
    double angoffset;               // angular offset between the angle of the perceived object and the angular center of the sector
    double maxperceivableaoffset;   // the maximum angular offset of a perceived object with respect to a sector
    double blobangext;              // angular extension of the perceived blob

    // initialize
    for(s=0; s < cro->nlasersensors; s++)
        mdist[s] = maxdist;

    // calculate the stimulation generated environmental objects
    for(ob=0; ob < nenvobjs; ob++)
    {
        switch(envobjs[ob].type)
          {
            // walls (we assume they are orthogonal with respect to the carthesian axes)
            case WALL:
                if (cro->x >= envobjs[ob].x && cro->x <= envobjs[ob].x2)
                    x = cro->x;
                 else
                    x = envobjs[ob].x;
                if (cro->y >= envobjs[ob].y && cro->y <= envobjs[ob].y2)
                    y = cro->y;
                 else
                    y = envobjs[ob].y;
                dx = (cro->x - x);
                dy = (cro->y - y);
                dist = sqrt(dx*dx+dy*dy) - cro->radius;
                angle = angv(x, y, cro->x, cro->y);
                angle = mangrelr(angle, cro->dir);
                if (angle > M_PI) angle -= PI2; // direction of the perceived object normalized between -M_PI and M_PI
                if (angle < -M_PI) angle += PI2;
                if (dist < maxdist)
                 {
                  // we compute the angle of the initial and final point of the wall
                  angle1 = angv(envobjs[ob].x, envobjs[ob].y, cro->x, cro->y);
                  angle1 = mangrelr(angle1, cro->dir);
                  if (angle1 > M_PI) angle1 -= PI2; // direction of the perceived object normalized between -M_PI and M_PI
                  if (angle1 < -M_PI) angle1 += PI2;
                  angle2 = angv(envobjs[ob].x2, envobjs[ob].y2, cro->x, cro->y);
                  angle2 = mangrelr(angle2, cro->dir);
                  if (angle2 > M_PI) angle2 -= PI2; // direction of the perceived object normalized between -M_PI and M_PI
                  if (angle2 < -M_PI) angle2 += PI2;

                  for(s=0, a1 = starta, a2 = starta + inta; s < cro->nlasersensors; s++, a1 += inta, a2 += inta)
                   {
                    // the pointer to the nearest point of wall is inside the sector
                    if (angle > a1 && angle < a2)
                     {
                      if (dist < mdist[s]) mdist[s] = dist;
                     }
                    else
                    {
                     // the first side point is inside the wall
                     if (anginside2(a1, angle1, angle2))
                      {
                        // calculates the length of the vector between the sensor and the object providing that the angular offset is smaller than 60 degrees
                        // for angular offset between 45 and 60 degrees the distance is increase linearly of a fraction of maxdist
                        angoffset = fabs(angdelta(a1, angle));
                        if (angoffset < (M_PI / 3.0))
                          {
                           ldist = dist * sin(angoffset);
                           cdist = sqrt((dist*dist)+(ldist*ldist));
                           if (angoffset > (M_PI / 4.0))
                              cdist += maxdist * ((angoffset - (M_PI / 4.0)) / (M_PI / 12.0));
                           if (cdist < mdist[s])
                             mdist[s] = cdist;
                           }
                       }
                     // the second side point is inside the wall
                     if (anginside2(a2, angle1, angle2))
                      {
                        // calculates the length of the vector between the sensor and the object providing that the angular offset is smaller than 60 degrees
                        // for angular offset between 45 and 60 degrees the distance is increase linearly of a fraction of maxdist
                        angoffset = fabs(angdelta(a2, angle));
                        if (angoffset < (M_PI / 3.0))
                          {
                           ldist = dist * sin(angoffset);
                           cdist = sqrt((dist*dist)+(ldist*ldist));
                           if (angoffset > (M_PI / 4.0))
                             cdist += maxdist * ((angoffset - (M_PI / 4.0)) / (M_PI / 12.0));
                           if (cdist < mdist[s])
                             mdist[s] = cdist;
                          }
                       }
                      }

                     }

                }
            break;
        // sampled cylinder
        case SAMPLEDSCYLINDER:
            dx = (cro->x - envobjs[ob].x);
            dy = (cro->y - envobjs[ob].y);
            dist = sqrt(dx*dx+dy*dy) - (cro->radius + envobjs[ob].r);
            angle = angv(envobjs[ob].x, envobjs[ob].y, cro->x, cro->y);
            angle = mangrelr(angle, cro->dir);
            if (angle > M_PI) angle -= PI2; // direction of the perceived object normalized between -M_PI and M_PI
            if (angle < -M_PI) angle += PI2;
            if (dist < maxdist)
             {
              // we compute the angle of the initial and final point of the wall
              angle1 = angv(envobjs[ob].x, envobjs[ob].y, cro->x, cro->y);
              angle1 = mangrelr(angle1, cro->dir);
              if (angle1 > M_PI) angle1 -= PI2; // direction of the perceived object normalized between -M_PI and M_PI
              if (angle1 < -M_PI) angle1 += PI2;
              angle2 = angv(envobjs[ob].x2, envobjs[ob].y2, cro->x, cro->y);
              angle2 = mangrelr(angle2, cro->dir);
              if (angle2 > M_PI) angle2 -= PI2; // direction of the perceived object normalized between -M_PI and M_PI
              if (angle2 < -M_PI) angle2 += PI2;

              if (dist < maxdist)
               {
                 for(s=0, a1 = starta, a2 = starta + inta; s < cro->nlasersensors; s++, a1 += inta, a2 += inta)
                    {
                      angoffset = fabs(angdelta(a1+(inta/2), angle));
                      maxperceivableaoffset = envobjs[ob].r / dist * (M_PI / 3);
                      if (maxperceivableaoffset > (M_PI / 3.0)) maxperceivableaoffset = M_PI / 3.0;  // in any case cannot exceeed 45 degrees
                      blobangext = envobjs[ob].r / dist;
                      if (blobangext > 1.0) blobangext = 1.0;
                      if (angoffset < maxperceivableaoffset)
                        {
                          cdist = dist + maxdist * pow((angoffset / maxperceivableaoffset), 2); // the perceived distance increases proportionally to the angular offset
                          cdist += maxdist * (1.0 - blobangext);                  // the perceived distance increases inversely to the angular extension of the perceived blob
                          if (cdist < mdist[s])
                             mdist[s] = cdist;
                        }
                    }
//printf("radius %.2f blobangext %.2f\n", envobjs[ob].r, blobangext);
               }

            }
        break;
            // default
            //default:
            //	printf("ERROR: undefined object type %d \n",envobjs[ob].type);
            break;
         }
    }
    // calculate the stimulation generated by the cylindrical body of other robots
//double absa;
    if (nrobots > 1)
      {
        for (r=0, ro=rob; r < nrobots; r++, ro++)
          {
            if (cro->idn != ro->idn)
             {
                dx = (cro->x - ro->x);
                dy = (cro->y - ro->y);
                dist = sqrt(dx*dx+dy*dy) - (cro->radius * 2.0);
                angle = angv(ro->x, ro->y, cro->x, cro->y);
                angle = mangrelr(angle, cro->dir);
                if (angle > M_PI) angle -= PI2; // direction of the perceived object normalized between -M_PI and M_PI
                if (angle < -M_PI) angle += PI2;
//printf("robot %d absa %.2f robotdir %.2f angle %.2f dist %.2f\n", cro->idn, absa, cro->dir, angle, dist);
                if (dist < maxdist)
                 {
                   for(s=0, a1 = starta, a2 = starta + inta; s < cro->nlasersensors; s++, a1 += inta, a2 += inta)
                      {
                        // the center of the perceived robot is inside the sector
                        if (anginside2(angle, a1, a2))
                         {
                          if (dist < mdist[s]) mdist[s] = dist;
//printf("center inside - robot %d angle %.2f sector %.2f %.2f dist %.2f\n", cro->idn, angle, a1, a2, dist);
                         }
                        else
                        {
                         if (dist < (cro->radius * 4.0)) // adjacent sectors can detect only in near proximity
                         {
                          // one of the side of the perceived robot is inside the maximum perceived angular offset of the sector
                          // when the distance is equal to the radius of the robot the angular offset is 45 degrees, longer the distance less the angularoffset
                          angoffset = angdelta(a1+(inta/2), angle);
                          maxperceivableaoffset = cro->radius / dist * (M_PI / 4);
                          if (maxperceivableaoffset > (M_PI / 4.0)) maxperceivableaoffset = M_PI / 4.0;  // cannot exceed 45 degrees
                          if (fabs(angoffset) < maxperceivableaoffset)
                            {
                             ldist = cro->radius * (angoffset / maxperceivableaoffset);
                             cdist = sqrt((dist*dist)+(ldist*ldist));
                             if (cdist < mdist[s])
                                mdist[s] = cdist;
//printf("side inside - robot %d angle %.2f sector %.2f %.2f dist %.2f angoffset %.2f maxperceivableoff %.2f ldist %.2f\n", cro->idn, angle, a1, a2, dist, angoffset, maxperceivableaoffset, ldist);
                            }
                           }
                        }


                      }
                 }
              }
           }
        }

    // normalize and copy on the imput vector
    for(s=0; s < cro->nlasersensors; s++)
        {
         *cro->csensors = (maxdist - mdist[s]) / maxdist;
         //printf("%.2f ", *cro->csensors);
         //*cro->csensors = *cro->csensors * *cro->csensors * *cro->csensors;
         cro->csensors++;
        }
    //printf("\n");
}

/*
 * generate the initial states
 * if mode != 0 regenerate only the trial initial conditions
 */
void createInitialStates(int mode)

{


}

/*
 * load an obstacle from a .sam file
 */
void
save_obstacle_new(char *filename, int  *objectcf, float **object)

{

   float  **ob;
   float  *o;
   int    *ocf;
   FILE   *fp;
   int    i,t,s;

   if ((fp = fopen(filename,"w!")) == NULL)
	 {
	    printf("I cannot open file %s", filename);
		fflush(stdout);
	  }
   // read configuration data: 0-nsensors, 1-nangles, 2-ndistances,
   //                          3-initdist,4-distinterval
   for (i=0, ocf=objectcf; i < 5; i++, ocf++)
     fprintf(fp,"%d ",*ocf);
   fprintf(fp,"\n");
	
   // save data
   for (t=0, ob=object; t < objectcf[2]; t++, ob++)
    {
     fprintf(fp,"TURN %d\n",t);
     for (i = 0, o = *ob; i < objectcf[1]; i++)
      {
       for (s=0; s < objectcf[0]; s++,o++)
        {
	     fprintf(fp,"%f ",*o);
        }
       fprintf(fp,"\n");
      }
    }
	
   fprintf(fp, "END\n");
   fclose(fp);
}


/*
 * initialize the environment, the robots, and the network
 * if mode != 0 regenerate the trial initial conditions
 * return the number of required parameters (-1 in case of failure)
 */
int initialize(const char* filename)

{
    
	// read parameters from the .ini file
	if (readEvoConfig(filename) == false)
		return(-1);
    
    nenvobjs = 1;            // number of objects
    worldx = 500.0;          // world x dimension
    worldy = 500.0;          // world y dimension
    
    envobjs = (struct envobjects *) malloc(nenvobjs * sizeof(struct envobjects));
    
    envobjs[0].type = WALL;
    envobjs[0].x = 0.0;
    envobjs[0].y = 0.0;
    envobjs[0].x2 = worldx;
    envobjs[0].y2 = 0.0;
            
    // create a khepera wheeled robot with 6 infrared, 1 bias, 2 motor outputs
	nrobots = 1;
    // set robot's properties
    rob = (struct robot *) malloc(nrobots * sizeof(struct robot));
    net = (struct evonet *) malloc(nrobots * sizeof(struct evonet));
	initRobot(rob, 0, MarXBot);
    rob->idn = 0;
    rob->color = 0;
    rob->camnsectors = 0;
    net->ninputs = 8;
    rob->nlasersensors = 8;
	//rob->nifsensors = 8;  // we set this to trigger the graphic rendering
    //net->ninputs += initKheperaInfraredSensor(rob, 8);
    rob->motorwheels = 0;
    rob->motorleds = 0;
    rob->motorwheelstype = 0;  // standard direct control (speed of the two wheels)
    net->noutputs = 0;
	initRobotSensors(rob, net->ninputs);
    // initialize robot's network
    initNet(net, filename, 0);
    resetNet(net);
    // set the id number of motors neurons
    rob->motorwheelsid = net->ninputs + net->nhiddens;
    rob->motorledsid = 0;
    
    nfreep = computeParameters(net);
    printf("Network parameters %d\n", nfreep);
    
    printf("end ini \n");
            
	return(nfreep);
}

/*
 * convert old files in the new float format
 * generate synthetic but realistic samples
 */
double evaluate(double* genotype, const int glen, int mode)

{
    
    int trial;
    int step;
    int n;
    double *cs;
    int s;
	
	float **object;
    float **ob;
    float *o;
	int   *objectcf;
	int i;
	int **iobject;
	
	// convert old files in the new format
	// NOTICE: floating poingt might be saved with "," instead of "." and should eventually be corrected
	/*
	objectcf = (int *) malloc(5 * sizeof(int));
	iobject = load_obstacle_old("ePuck-WALL.sam", objectcf);
	save_obstacle("epuck-wall.sample", iobject, objectcf);
	iobject = load_obstacle_old("ePuck-SMALL.sam", objectcf);
	save_obstacle("epuck-scylinder.sample", iobject, objectcf);
	iobject = load_obstacle_old("ePuck-LIGHT.sam", objectcf);
	save_obstacle("epuck-light.sample", iobject, objectcf);
	*/
	
    // allocate space and initialize
    objectcf = (int *) malloc(5 * sizeof(int));
    objectcf[0] = 8;		// sensors
    objectcf[1] = 180;	// nangles
    objectcf[2] = 50;	// ndistances
    objectcf[3] = 4;		// init distance
    objectcf[4] = 2;		// dist interval
	
	
   object = (float **) malloc(objectcf[2] * sizeof(float *));
   ob = object;
   for (i=0; i<objectcf[2]; i++,ob++)
	   *ob = (float *)malloc((objectcf[0] * objectcf[1]) * sizeof(float));
	

    for (n=0; n < 3; n++)
    {
     if (n == 0) // wall
      {
       envobjs[0].type = WALL;
       envobjs[0].x = worldx;
       envobjs[0].y = 0.0;
       envobjs[0].x2 = worldx;
       envobjs[0].y2 = worldy;
      }
     if (n == 1) // large cylinder
      {
        envobjs[0].type = SAMPLEDSCYLINDER;
        envobjs[0].x = worldx;
        envobjs[0].y = worldy/2.0;
        envobjs[0].r = rob->radius;
      }
     if (n == 2) // small cylinder
      {
        envobjs[0].type = SAMPLEDSCYLINDER;
        envobjs[0].x = worldx;
        envobjs[0].y = worldy/2.0;
        envobjs[0].r = rob->radius / 2.0;
      }
	for (trial=0, ob = object; trial < objectcf[2]; trial++, ob++)
    {
        rob->dir = 0.0;
        if (n == 0)
          rob->x = worldx - rob->radius - objectcf[3] - (trial * objectcf[4]);
         else
          rob->x = worldx - (rob->radius + envobjs[0].r) - objectcf[3] - (trial * objectcf[4]);
        rob->y = worldy / 2.0;
        for (step=0, o = *ob; step < 180; step++)
        {
            rob->csensors = rob->sensors;
            updateSampleLaserDistanceSensor(rob);
			//updateNet(net, rob->sensors, genotype);
            for (s=0, cs = rob->sensors; s < 8; s++, cs++, o++)
			  {
			    net->act[s] = *cs;
                *o = pow(*cs, 2);  // non-linear transformation of distances
			  }
            //updateKheperaInfraredSensor(rob);
            printf("%d %d %.2f %.2f %.2f) ", trial, step, rob->x, rob->y, rob->dir);
            for (s=0, cs = rob->sensors; s < 8; s++, cs++)
              printf("%.2f ", *cs);
            printf("\n");
            rob->dir -= M_PI / 90.0;
            
#ifdef __cplusplus

            if (renderrobotworld)
               update_world();		// update graphic rendering
        }
#endif
        }
         if (n == 0) save_obstacle_new("marxbot-wall.sample", objectcf, object);
         if (n == 1) save_obstacle_new("marxbot-cylinder.sample", objectcf, object);
         if (n == 2) save_obstacle_new("marxbot-scylinder.sample", objectcf, object);
        }
	


    return(0.0);
    
}


