/*
 * discrim.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "robot-env.h"
#include "evonet.h"
#include "utilities.h"

/*
 * read parameters from the configuration file
 */
bool readEvoConfig(const char* filename)
{
    char *s;
    char buff[1024];
    char name[1024];
    char value[1024];
    char *ptr;
    int section;  // 0=before the section 1=in the section 2= after the section

    section = 0;

    FILE* fp = fopen(filename, "r");
    if (fp != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp) != NULL)
        {

            // check whether the relevant section start or end
            if (buff[0] == '[')
            {
              if (section == 1)
                {
                    section = 2;
                    continue;
                }
              if ((section == 0) && (strncmp(buff, "[EXP]",5)==0))
              {
                section = 1;
                continue;
              }
            }

            if (section == 1)
            {

            //Skip blank lines and comments
            if (buff[0] == '\n' || buff[0] == '#' || buff[0] == '/')
            continue;

            //Parse name/value pair from line
            s = strtok(buff, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, name);

            s = strtok(NULL, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, value);

            // Copy into correct entry in parameters struct
            if (strcmp(name, "trials")==0)
            trials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "ttrials")==0)
            ttrials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steps")==0)
            steps = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steptimelength")==0)
            steptimelength = (int)strtol(value, &ptr, 10);
            else printf("WARNING: Unknown parameter %s in section [EXP] of file %s \n", name, filename);
         }
        }
        fclose (fp);
        if (section == 0)
           printf("WARNING: Missing section [EXP] in file %s \n", filename);

        printf("Trials %d TTrials %d Steps %d\n", trials, ttrials, steps);
        return(true);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
        return(false);
    }
}



/*
 * initialize the environment
 */
void
initEnvironment()

{

	int cobj=0;
	
	nenvobjs = 5;			// number of objects
	worldx = 1000.0;			// world x dimension (it was 400)
	worldy = 1000.0;			// world y dimension (it was 400)
	
	envobjs = (struct envobjects *) malloc(nenvobjs * sizeof(struct envobjects));
	
	envobjs[cobj].type = SAMPLEDSCYLINDER;
	envobjs[cobj].x = worldx/2.0;
	envobjs[cobj].y = worldy/2.0;
	envobjs[cobj].r = 40.0;  // it was 12.5
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = 0.0;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = 0.0;
	envobjs[cobj].y2 = worldy;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = worldx;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = worldy;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = worldy;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = worldy;
	cobj++;
	
	if (cobj > nenvobjs)
		{
			printf("ERROR: you should allocate more space for environmental objects");
			fflush(stdout);
			exit;
		}
	
}


/*
 * generate the initial states
 * if mode != 0 regenerate only the trial initial conditions
 */
void createInitialStates(int mode)

{
    
    int s;
    double *is;
	int numtrials;
	double tx, ty, x, y, dist;
	int attempts;

	if (mode == 0)
	  numtrials = trials + ttrials;
	 else
	  numtrials = trials;

    for (s=0, is = initialstates; s < numtrials; s++)
    {
        // generate the position of the target and the direction of the robot
        tx = *is = RNG->getDouble(rob->radius + 10.0, worldx - (rob->radius + 10.0)); is++;
        ty = *is = RNG->getDouble(rob->radius + 10.0, worldy - (rob->radius + 10.0)); is++;
        *is = RNG->getDouble(0.0, PI2); is++;
        // generate the position of the robot by checking that it is not over the target
        dist = 0.0; attempts = 0;
        while(dist < 120 && attempts < 100)
		{
		 x = RNG->getDouble(50.0, worldx - 100.0);
         y = RNG->getDouble(50.0, worldy - 100.0);
         dist = sqrt(((x - tx) * (x - tx)) + ((y - ty) * (y - ty)));
		 attempts++;
		}
        //printf("trial % d dist %.2f \n", s, dist);
		*is = x; is++;
		*is = y; is++;
	}

}


/*
 * initialize the environment, the robots, and the network
 * if mode != 0 regenerate the trial initial conditions
 * return the number of required parameters (-1 in case of failure)
 */
int initialize(const char* filename)

{
    
	// default values
    trials = 1;
    ttrials = 0;
    steps = 100;
	
	// read parameters from the .ini file
	if (readEvoConfig(filename) == false)
		return(-1);
	
    // create and configure the environment
    initEnvironment();
	
    // create a khepera wheeled robot with 6 infrared, 1 bias, 2 motor outputs
	nrobots = 1;
    // set robot's properties
    rob = (struct robot *) malloc(nrobots * sizeof(struct robot));
    net = (struct evonet *) malloc(nrobots * sizeof(struct evonet));
	initRobot(rob, 0, MarXBot);
    rob->idn = 0;
    rob->color = 0;
    // initialize sensors
    net->ninputs = 0;
    net->ninputs += initKheperaInfraredSensor(rob, 6); // the 6 frontal infrared sensors only
    net->ninputs += initBiasSensor(rob);
    rob->camnsectors = 0; // THIS IS NEEDED TEMPORARITY TO AVOID THAT THE PAINTER TRY TO PRINT BLOBS
    // initialize motors
    rob->motorwheels = 2;
    rob->motorleds = 0;
    rob->motorwheelstype = 0;  // standard direct control (speed of the two wheels)
    net->noutputs = rob->motorwheels;
    // initialize the input vector
	initRobotSensors(rob, net->ninputs);
    // initialize robot's network
    initNet(net, filename, 0);
    resetNet(net);
    // set the id number of motors neurons
    rob->motorwheelsid = net->ninputs + net->nhiddens;
    rob->motorledsid = 0;

    // generate initial states
    // (initial positions/orientations of robots and positions of foods area)
    initialstates = (double *) malloc(((trials+ttrials)*5) * sizeof(double));
    createInitialStates(0);
    
    nfreep = computeParameters(net);
    printf("Network parameters %d\n", nfreep);
	
	return(nfreep);
}

/*
 * evaluate a genotype for capability to perform the discrimination task
 * return the total fitness
 * when mode=1 evaluate a genotype on the test trials
 */
double evaluate(double* genotype, const int glen, int mode)

{
    
    double *is;
    int trial;
	int numtrials;
    int step;
    double fitness;
    double dx, dy, dist;
	
    // evaluation loop
	if (glen != nfreep)
	  {
	    printf("WARNING: The length of the parameter vector received by discriminate (%d) does not match the number of parameters required (%d)\n", glen, nfreep);
	  }
    fitness = 0.0;
	if (mode == 0)
	   {
	    is = initialstates;
		numtrials = trials;
		}
	  else
	   {
	    is = (initialstates + (trials * 3));
		numtrials = ttrials;
	   }
	for (trial=0, is = initialstates; trial < numtrials; trial++)
    {
        resetNet(net);
        envobjs[0].x = *is; is++;
        envobjs[0].y = *is; is++;
        rob->dir = *is; is++;
        rob->x = *is; is++;
        rob->y = *is; is++;
        for (step=0; step < steps; step++)
        {
            rob->csensors = rob->sensors;
            updateKheperaInfraredSensor(rob);
            updateBiasSensor(rob);
            updateNet(net, rob->sensors, genotype);
            if (updateRobot(rob, net) != 0)
			  break;
#ifdef __cplusplus

			if (renderrobotworld)
               update_world();		// update graphic rendering
#endif
			
            //printf("step %d) pos %.2f %.2f %.2f  %.2f]\n", step, rob->x, rob->y, rob->dir, fitness);
            //for (int n=0; n < net->nneurons; n++)
            //    printf("%.3f ", net->act[n]);
            //printf("\n");
			
			// update fitness
            dx = envobjs[0].x - rob->x;
            dy = envobjs[0].y - rob->y;
            dist = sqrt((dx*dx)+(dy*dy));
            //printf("dist %.2f \n", dist);
            if (dist < (rob->radius + envobjs[0].r + 50))
			  fitness += 1;

            }
        }
	
    return(fitness / (double) numtrials);
    
}


