#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include "evonet.h"

extern int trials;                        // number of trials
extern int ttrials;                       // number of testing trials

struct evonet *net;                  // list of networks
double *inputpattern;                // the current input pattern
double *initialstates;               // robots initial positions


int steps=1000;                      // number of steps for each trial
int nfreep;                          // number of free parameters
bool fixed = false;                  // whether we use fixed initial position

// Cart-pole
#define MASSPOLE_1 0.1
#define LENGTH_1 0.5
double MASSPOLE_2 = 0.01;
double LENGTH_2 = 0.05;
#define MUP 0.000002
#define MUC 0.0005
#define GRAVITY -9.8
#define MASSCART 1.0
#define FORCE_MAG 10.0
#define TAU 0.01
#define NUM_INPUTS 4
#define NUM_STATES 6
double state[NUM_STATES];
// Ranges for state variables
double stateRanges[NUM_STATES];
// Constant value representing an angle of 36°
const double thirty_six_degrees = 0.628329;


static double mrand(double range)
{
    return(((double)rand() / (double)RAND_MAX) * (range * 2) - range);
}


/*
 * Check whether the cart is located within the 4.8m area
 * and whether the angle of the poles does not exceed 36 degrees
 */
static bool outsideBounds()
{

    return
    state[0] < -2.4          ||
    state[0] > 2.4           ||
    state[2] < -0.628329     ||
    state[2] > 0.628329      ||
    state[4] < -0.628329     ||
    state[4] > 0.628329;
}

static void step(double acton, double *st, double *derivs)
{
    double force,costheta_1,costheta_2,sintheta_1,sintheta_2,
          gsintheta_1,gsintheta_2,temp_1,temp_2,ml_1,ml_2,fi_1,fi_2,mi_1,mi_2;

    force =  (acton - 0.5) * FORCE_MAG * 2;
    costheta_1 = cos(st[2]);
    sintheta_1 = sin(st[2]);
    gsintheta_1 = GRAVITY * sintheta_1;
    costheta_2 = cos(st[4]);
    sintheta_2 = sin(st[4]);
    gsintheta_2 = GRAVITY * sintheta_2;

    ml_1 = LENGTH_1 * MASSPOLE_1;
    ml_2 = LENGTH_2 * MASSPOLE_2;
    temp_1 = MUP * st[3] / ml_1;
    temp_2 = MUP * st[5] / ml_2;
    fi_1 = (ml_1 * st[3] * st[3] * sintheta_1) +
           (0.75 * MASSPOLE_1 * costheta_1 * (temp_1 + gsintheta_1));
    fi_2 = (ml_2 * st[5] * st[5] * sintheta_2) +
           (0.75 * MASSPOLE_2 * costheta_2 * (temp_2 + gsintheta_2));
    mi_1 = MASSPOLE_1 * (1 - (0.75 * costheta_1 * costheta_1));
    mi_2 = MASSPOLE_2 * (1 - (0.75 * costheta_2 * costheta_2));

    derivs[1] = (force + fi_1 + fi_2)
                 / (mi_1 + mi_2 + MASSCART);

    derivs[3] = -0.75 * (derivs[1] * costheta_1 + gsintheta_1 + temp_1)
                 / LENGTH_1;
    derivs[5] = -0.75 * (derivs[1] * costheta_2 + gsintheta_2 + temp_2)
                  / LENGTH_2;

}

static void rk4(double f, double y[], double dydx[], double yout[])
{

    int i;

    double hh,h6,dym[6],dyt[6],yt[6];


    hh=TAU*0.5;
    h6=TAU/6.0;
    for (i=0;i<=5;i++) yt[i]=y[i]+hh*dydx[i];
    step(f,yt,dyt);
    dyt[0] = yt[1];
    dyt[2] = yt[3];
    dyt[4] = yt[5];
    for (i=0;i<=5;i++) yt[i]=y[i]+hh*dyt[i];
    step(f,yt,dym);
    dym[0] = yt[1];
    dym[2] = yt[3];
    dym[4] = yt[5];
    for (i=0;i<=5;i++) {
        yt[i]=y[i]+TAU*dym[i];
        dym[i] += dyt[i];
    }
    step(f,yt,dyt);
    dyt[0] = yt[1];
    dyt[2] = yt[3];
    dyt[4] = yt[5];
    for (i=0;i<=5;i++)
        yout[i]=y[i]+h6*(dydx[i]+dyt[i]+2.0*dym[i]);
}

/*
 * Apply the chosen torque and compute the new state of the cart and of the poles
 * by calling the functions step() and rk4()
 */
static void performacton(double output, int stepnum)
{

  int i;
  double  dydx[6];

  const bool RK4=true; //Use the Runge-Kutta 4th order integration method
  const double EULER_TAU= TAU/4;

  /*--- Apply acton to the simulated cart-pole ---*/

  if(RK4){
    for(i=0;i<2;++i){
      dydx[0] = state[1];
      dydx[2] = state[3];
      dydx[4] = state[5];
      step(output,state,dydx);
      rk4(output,state,dydx,state);
    }
  }
  else{
    for(i=0;i<8;++i){
      step(output,state,dydx);
      state[0] += EULER_TAU * dydx[0];
      state[1] += EULER_TAU * dydx[1];
      state[2] += EULER_TAU * dydx[2];
      state[3] += EULER_TAU * dydx[3];
      state[4] += EULER_TAU * dydx[4];
      state[5] += EULER_TAU * dydx[5];
    }
  }

}


/*
 * read parameters from the configuration file
 */
static bool readConfig(const char* filename)
{
    char *s;
    char buff[1024];
    char name[1024];
    char value[1024];
    char *ptr;
	
    FILE* fp = fopen(filename, "r");
    if (fp != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp) != NULL)
        {
            //Skip blank lines and comments
            if (buff[0] == '\n' || buff[0] == '#' || buff[0] == '['  || buff[0] == '/')
            continue;
            
            //Parse name/value pair from line
            s = strtok(buff, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, name);
            
            s = strtok(NULL, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, value);
            
            // Copy into correct entry in parameters struct
            if (strcmp(name, "verbose")==0)
            verbose = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "trials")==0)
            trials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "ttrials")==0)
            ttrials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steps")==0)
            steps = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "polelen2")==0)
            LENGTH_2 = strtod(value, &ptr);
            else if (strcmp(name, "polemass2")==0)
            MASSPOLE_2 = strtod(value, &ptr);
            else if (strcmp(name, "fixedPositions")==0)
            fixed = (bool)strtol(value, &ptr, 10);
        }
        fclose (fp);
        if (verbose >= 1)
        {
            printf("trials %d ttrials %d Fixed %d PoleL %.2f %.2f PoleM %.2f %.2f\n", trials, ttrials, fixed, LENGTH_1, LENGTH_2, MASSPOLE_1, MASSPOLE_2);
        }
        return(true);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
        return(false);
    }
}

/*
 * save a genotype in a file
 */
void savegenotype(char* filename, double* genotype, const int glen, int mode)

{
    
    double *gene;
    int g;
    FILE *fp;
    
    fp = fopen(filename,"w!");
    for (g=0, gene=genotype; g < glen; g++, gene++)
    {
        fprintf(fp, "%lf ", *gene);
    }
    fprintf(fp, "\n");
    fclose(fp);
    
}

/*
 * generate or regenerate the initial training and testing states
 * when mode is 1 re-generate only the training states
 */
void createInitialStates(int mode)
{
    int t, s;
    double *st;
	int ntrials;

    stateRanges[0] = 1.944;
    stateRanges[1] = 1.215;
    stateRanges[2] = 0.10472;
    stateRanges[3] = 0.135088;
    stateRanges[4] = stateRanges[2];
    stateRanges[5] = stateRanges[3];

    st = initialstates;
    
    if (fixed)
    {
	// Fixed initial states
        for (t = 0; t < trials; t++)
        {
            switch (t)
	    {
		case 0:
		    *st = -stateRanges[0];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		case 1:
		    *st = stateRanges[0];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		case 2:
		    *st = 0.0;
		    st++;
		    *st = -stateRanges[1];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		case 3:
		    *st = 0.0;
		    st++;
		    *st = stateRanges[1];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		case 4:
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = -stateRanges[2];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		case 5:
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = stateRanges[2];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		case 6:
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = -stateRanges[3];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		case 7:
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    *st = stateRanges[3];
		    st++;
		    *st = 0.0;
		    st++;
		    *st = 0.0;
		    st++;
		    break;
		default:
		    printf("ERROR - Invalid value for trial %d!!!", t);
		    break;
	    }
        }
    }
    else
    {
		// Randomly varying initial states
		// Test trials are located after training trials
		if (mode == 0)
			ntrials = trials + ttrials;
		   else
		    ntrials = trials;
        for (t = 0; t < ntrials; t++)
        {
            for (s = 0; s < NUM_STATES; s++, st++)
            {
                *st = mrand(stateRanges[s]);
            }
        }
    }
}

int initialize(const char* filename)
{
    // read parameters from the configuration file
    bool ret = readConfig(filename);
    if (!ret)
		printf("ERROR: unable to read parameters from file %s!\n", filename);
    
    // create a and initialize network
    net = (struct evonet *) malloc(1 * sizeof(struct evonet));
	net->ninputs = 4; // x pos, poles angles, and bias
	net->noutputs = 1;
    initNet(net, filename, 0);
    resetNet(net);
    nfreep = computeParameters(net);
    printf("Network require %d parameters\n", nfreep);
    
    // allocate space for the input pattern
    inputpattern = (double *) malloc(NUM_INPUTS * sizeof(double));
    
    // Allocate space for and create initial states
    initialstates = (double*)malloc(sizeof(double) * (NUM_STATES * (trials + ttrials)));
    createInitialStates(0);
    
    // Sanity check
    if (fixed && (trials != 8))
        printf("ERROR: if you want to use fixed initial states the number of trials should be 8 and not %d \n", trials);
    if ((net->ninputs != 3 && net->ninputs != 4) || net->noutputs != 1)
        printf("ERROR: the network should have 3-4 input neurons and 1 output neurons \n");
	return (nfreep);
}

/*
 * evaluate for n. trials (mode should be set to 0)
 * receive the genotype and the number of parameters
 * when mode is set to 1, it evaluate for n. ttrials on the test set
 */
double evaluate(double* genotype, const int glen, int mode)
{
    int t;
    int i;
    double fit;
    double *p;
    int step;
    double output;
    int n;
    double *a;
    double *in;
    double *st;
    int s;
	int ntrials;
	int it;
    double* freep;
   
    freep = genotype;
	
    /* Print number of parameters and genotype
    if (verbose >= 2)
    {
        printf("num parameters %d\n", computeParameters(net));
        for (i = 0, p = freep; i < glen; i++, p++)
            printf("%.1f ", *p);
        printf("\n");
        fflush(stdout);
    }*/
	
	// sanity check
	if (glen != computeParameters(net))
	{
	   printf("ERROR: genotype length (%d) does not match the parameters required by doublepole.c (%d)\n", glen, computeParameters(net));
	   fflush(stdout);
	   return(-9999.0);
	}
	
    // evaluate for multiple trials
	if (mode == 0)
	  {
	   ntrials = trials;
	   it = 0;
	  }
	 else
	  {
	   ntrials = ttrials;
	   it = trials;
	  }
    for (t = 0, fit = 0.0; t < ntrials; t++)
    {
	resetNet(net);
	
	// Initialize the state of the cart
	st = (initialstates + ((t + it) * NUM_STATES));
	for (s = 0; s < NUM_STATES; s++, st++)
	    state[s] = *st;
         
        //Trial loop
        step = 0;
        while (step < steps)
        {
            // set the input
            in = inputpattern;
            *in = state[0] / 4.8;
            in++;
            *in = state[2]  / 0.52;
            in++;
            *in = state[4] / 0.52;
            in++;
            *in = 0.5;
             
            // actvate the net
            updateNet(net, inputpattern, genotype);
             
            /* save the actvations of the neurons
            if (verbose >= 2)
            {
                FILE* fp = fopen("activation.txt", "a");
                fprintf(fp, "%d  %d ",t, step);
                for(n=0, a=net->act; n < (net->ninputs + net->nhiddens + net->noutputs); n++,a++)
                    if (n < net->ninputs)
                        fprintf(fp, "%.3f ",inputpattern[n]);
                    else
                        fprintf(fp, "%.3f ",*a);
                fprintf(fp, "\n");
                fclose(fp);
            }
			*/
             
            // Get the network output
            output = getOutput(net, 0);
             
            // Perform acton (i.e., the new state)
            performacton(output,steps);
             
            // terminate prematurely in case of failure
            if (outsideBounds())
                break;
            
            step++;
        }
         
	fit += (double) step / (double) steps;
          
    }
    
    return( fit / (double) ntrials);
}
