/*
 * discrim.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <locale.h>
#include "robot-env.h"
#include "evonet.h"
#include "utilities.h"

double **actions = NULL;    // matrix used to store motor states over time
double complexlev[4];       // summed intergrals of speeds averaged over 10 steps (1s)

/*
 * read parameters from the configuration file
 */
bool readEvoConfig(const char* filename)
{
    char *s;
    char buff[1024];
    char name[1024];
    char value[1024];
    char *ptr;
	
    
    FILE* fp = fopen(filename, "r");
    if (fp != NULL)
    {
        // Read lines
        while (fgets(buff, 1024, fp) != NULL)
        {
            //Skip blank lines and comments
            if (buff[0] == '\n' || buff[0] == '#' || buff[0] == '['  || buff[0] == '/')
            continue;
            
            //Parse name/value pair from line
            s = strtok(buff, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, name);
            
            s = strtok(NULL, " = ");
            if (s == NULL)
            continue;
            else
            copyandclear(s, value);
            
            // Copy into correct entry in parameters struct
			if (strcmp(name, "nrobots")==0)
            nrobots = (int)strtol(value, &ptr, 10);
			if (strcmp(name, "trials")==0)
            trials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "ttrials")==0)
            ttrials = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steps")==0)
            steps = (int)strtol(value, &ptr, 10);
            else if (strcmp(name, "steptimelength")==0)
            steptimelength = (int)strtol(value, &ptr, 10);
		 }
        fclose (fp);
		
		printf("Robots %d Trials %d TTrials %d Steps %d\n", nrobots, trials, ttrials, steps);
        return(true);
    }
    else
    {
        printf("ERROR: unable to open file %s\n", filename);
        fflush(stdout);
        return(false);
    }
}

/*
 * initialize the environment
 */
void
initEnvironment()

{

	int cobj=0;
	
    nenvobjs = 4;	// total number of objects
    worldx = 3000;
    worldy = 3000;
	int f;
	
	envobjs = (struct envobjects *) malloc(nenvobjs * sizeof(struct envobjects));
	
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = 0.0;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = 0.0;
	envobjs[cobj].y2 = worldy;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = worldx;
	envobjs[cobj].y = 0.0;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = worldy;
	cobj++;
	envobjs[cobj].type = WALL;
	envobjs[cobj].x = 0.0;
	envobjs[cobj].y = worldy;
	envobjs[cobj].x2 = worldx;
	envobjs[cobj].y2 = worldy;
	cobj++;

	if (cobj > nenvobjs)
		{
			printf("ERROR: you should allocate more space for environmental objects");
			fflush(stdout);
		}
	
}


/*
 * generate the initial states
 * we do not need them since the robots start always from the same initial position and orientation
 */
void createInitialStates(int mode)

{

}

/* 
 * compute the complexity of robots' behavior
 */
void compute_complexity(int step, bool stoptrial)

{

	struct evonet *ne;
	double **ac;
	double *a;
	int s;
	int m;
	int av;
    double avs;
    double oldavs;
	
	// each step store the action state
    ne = net;
	ac = actions;
	a = *ac;
	*(a + step) = ne->act[ne->ninputs+ne->nhiddens];
    
	ac = (actions + 1);
	ne = net;
	a = *ac;
	*(a + step) = ne->act[ne->ninputs+ne->nhiddens+1];
    
	ac = (actions + 2);
	ne = (net + 1);
	a = *ac;
	*(a + step) = ne->act[ne->ninputs+ne->nhiddens];
    
	ac = (actions + 3);
	ne = (net + 1);
	a = *ac;
	*(a + step) = ne->act[ne->ninputs+ne->nhiddens+1];
	
	// at the last step we calculate the moving average
	// and speed variation providing that we have at least 100 steps
    for (m=0; m < 4; m++)
        complexlev[m] = 0.0;
	if (step > 100 && (stoptrial || step == (steps - 1)))
	 {
	  for (m=0, ac = actions; m < 4; m++, ac++)
	   {
        // compute the summed variation of average speed, speed is averaged over 10 steps
        // the first and the last set of 5 steps are excluded from the calculation of variation
	    for(s=0, a = *ac; s < (step - 10); s++)
	     {
		  for(av=0, avs = 0.0; av < 10; av++)
			  avs += *(a + (s + av));
          if (s > 0)
            complexlev[m] += fabs(avs - oldavs);
          oldavs = avs;
	     }
	   }
      for (m=0; m < 4; m++)
         complexlev[m] = complexlev[m] / (double) (step - 10);
      complexlev[0] = (complexlev[0] + complexlev[1] / 2.0);  // average predators motors
      complexlev[1] = (complexlev[2] + complexlev[3] / 2.0);  // average prey motors
	 }
     else
     {
       complexlev[0] = complexlev[1] = -1.0;
     }
}


/*
 * initialize the environment, the robots, and the network
 * if mode != 0 regenerate the trial initial conditions
 * return the number of required parameters (-1 in case of failure)
 */
int initialize(const char* filename)

{
    
    struct robot *ro;
    struct evonet *ne;
	int r;
    double **ac;
    int m;

	// default values
    trials = 1;
    ttrials = 1;
    steps = 100;
	
	// set USA local conventions
	setlocale( LC_ALL, "en-US" );
	// read parameters from the .ini file
	if (readEvoConfig(filename) == false)
		return(-1);

    // create and configure the environment
    initEnvironment();
    
    // allocate the matrix used to compute behavioral complexity
    // it is used to store the state of the 2 motors for pred and prey robots over steps
    actions = (double **) malloc(4 * sizeof(double *));
    for (m=0, ac=actions; m < 4; m++, ac++)
        *ac = (double *) malloc(steps * sizeof(double));

    // create the robots, the networks, and initialize the sensors
    rob = (struct robot *) malloc(nrobots * sizeof(struct robot));
	net = (struct evonet *) malloc(nrobots * sizeof(struct evonet));
    // initialize the robots and the networks
	for (r=0, ro=rob, ne=net; r < nrobots; r++, ro++, ne++)
	 {
	  initRobot(ro, r, MarXBot);
      // set robot's properties
      if (r == 0)
         ro->color = 1;  // predators and red
        else
         ro->color = 3;  // prey are green
      ro->camnsectors = 0; //
      // initialize sensors
      ne->ninputs = 0;
      ne->ninputs += initInfraredSensor(ro, 8, ne->ninputs);
      ne->ninputs += initCameraPPSensor(ro);
      ne->ninputs += initGroundGradSensor(ro);
	  ne->ninputs += initTimeSensor(ro);
      ne->ninputs += initBiasSensor(ro);
      // initialize motors
      ro->motorwheels = 2;
      ro->motorwheelstype = 0; // direct wheel control
      ro->motorleds = 0;
      ne->noutputs = ro->motorwheels;
      // initialize the robots
	  initRobotSensors(ro, ne->ninputs);
	  if (r == 0)
         ro->maxSpeed = 500.0 * 0.80;  // predator is 20% slower
		else
		 ro->maxSpeed = 500.0;
      // initialize robot's network
      initNet(ne, filename, r);
      resetNet(ne);
      // set the id number of motors neurons
      ro->motorwheelsid = ne->ninputs + ne->nhiddens;
      ro->motorledsid = ne->ninputs + ne->nhiddens;
	 }
	
	nfreep = computeParameters(net);
	printf("Network parameters %d\n", nfreep);
	
	return(nfreep);
}

/*
 * evaluate a genotype for capability to perform the discrimination task
 * return the total fitness
 * when mode=1 evaluate a genotype on the test trials
 */

double evaluate2(double* genotype1, double* genotype2, const int glen, int phase)

{
	
	int r,f,s;
    double *is;
	int numtrials;
    int step;
    double fitness;
	struct robot *ro;
	struct evonet *ne;
	bool stoptrial;
    double oldx, oldy; // updated every 2s
	double dx, dy;
	double cdist;
    int *rbd;
    int x, y;
	
	
	// evaluation loop
	if (glen != nfreep)
	    {
	    printf("WARNING: The length of the parameter vector received by discriminate (%d) does not match the number of parameters required (%d)\n", glen, nfreep);
	    }

	// default fitness (valid when the two robots do not collide during the trial)
	if (phase == 0)
		fitness = 0.0;
	   else
	    fitness = 1.0;
	
	stoptrial = false;
    
	for (r=0, ro=rob, ne=net; r < nrobots; r++, ro++, ne++)
		  {
           resetNet(ne);
			for(s=0,ro->csensors=ro->sensors; s < ne->ninputs; s++, ro->csensors++)
		      *ro->csensors = 0.0;
		  }
	// initialize robots position, orientation, and energy level
	for (r=0, ro=rob; r < nrobots; r++, ro++)
	     {
           if (r == 0)
           {
             ro->x = worldx / 3.0;
             ro->y = worldy / 2.0;
             ro->dir = 0.0;
           }
           else
           {
             ro->x = worldx / 3.0 * 2.0;
             ro->y = worldy / 2.0;
             ro->dir = M_PI;
           }
	     }

	for (step=0; step < steps; step++)
        {
          for (r=0, ro=rob, ne=net; r < nrobots; r++, ro++, ne++)
		  {
			// update sensors
            ro->csensors = ro->sensors;
            updateInfraredSensor(ro);
            updateCameraPPSensor(ro);
            updateGroundGradSensor(ro);
			updateTimeSensor(ro, step, steps);
			updateBiasSensor(ro);
			// update network
			if (r == 0)
			  updateNet(ne, ro->sensors, genotype1);
			 else
			  updateNet(ne, ro->sensors, genotype2);
			// update the robot
			if (updateRobot(ro, ne) > 0)
			  {
				stoptrial=true;
				if (phase == 0)
				  fitness = 1.0 - ((double) step / (double) steps);
				 else
				  fitness = ((double) step / (double) steps);
			  }
		  }
			
#ifdef __cplusplus
          if (renderrobotworld)
		    {
			  update_world();		// update graphic rendering
			}
#endif
         //compute_complexity(step, stoptrial);
		 if (stoptrial)
		   break;
			
	   }
    
    // printf("(%.2f %.2f)\n", complexlev[0] + complexlev[1], complexlev[2] + complexlev[3]);
    return(fitness);
    
}


/*
double evaluate(double* genotype, const int glen, int mode)

{
	
	int r,f,s;
    double *is;
    int trial;
	int numtrials;
    int step;
    double fitness;
	double tfitness;
	struct robot *ro;
	struct evonet *ne;
	bool stoptrial;
    double oldx, oldy; // updated every 2s
	double dx, dy;
    int foodarea[100];
	int collectedfood[100];
	double cdist;
    int *rbd;
	int coll;
    int x, y;
	
	
    // evaluation loop
	if (glen != nfreep)
	  {
	    printf("WARNING: The length of the parameter vector received by discriminate (%d) does not match the number of parameters required (%d)\n", glen, nfreep);
	  }

    fitness = 0.0;
    for (trial=0; trial < trials; trial++)
    {
	    // the robots initially do not carry food
		tfitness = 0.0;
		stoptrial = false;

	    for (r=0, ro=rob, ne=net; r < nrobots; r++, ro++, ne++)
		  {
           resetNet(ne);
			for(s=0,ro->csensors=ro->sensors; s < ne->ninputs; s++, ro->csensors++)
		      *ro->csensors = 0.0;
		  }
        // initialize robots position, orientation, and energy level
	    for (r=0, ro=rob; r < nrobots; r++, ro++)
	     {
           if (r == 0)
           {
             ro->x = worldx / 3.0;
             ro->y = worldy / 2.0;
             ro->dir = 0.0;
           }
           else
           {
             ro->x = worldx / 3.0 * 2.0;
             ro->y = worldy / 2.0;
             ro->dir = M_PI;
           }
	     }

        for (step=0; step < steps; step++)
        {
          for (r=0, ro=rob, ne=net; r < nrobots; r++, ro++, ne++)
		  {
			// update sensors
            ro->csensors = ro->sensors;
            updateKheperaInfraredSensor(ro);
            updateCameraPPSensor(ro);
            updateGroundGradSensor(ro);
			updateBiasSensor(ro);
			// update network
			updateNet(ne, ro->sensors, genotype);
			// update the robot
			coll = updateRobot(ro, ne);
			if (coll == 2)
			  {
				stoptrial=true;
				tfitness += 1.0 - ((double) step / (double) steps);
			  }
		  }
			


#ifdef __cplusplus
          if (renderrobotworld) update_world();		// update graphic rendering
#endif
		 if (stoptrial)
		   break;
			
	   }
	   fitness += tfitness;
		
	}
	

	
    return(fitness / trials);
    
}
*/

